#ifndef DRV_TAPI_H
#define DRV_TAPI_H

/******************************************************************************

                              Copyright (c) 2009
                            Lantiq Deutschland GmbH
                     Am Campeon 3; 85579 Neubiberg, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file drv_tapi.h
   Contains TAPI functions declaration und structures.
*/


/* ============================= */
/* Includes                      */
/* ============================= */
#include "drv_api.h"
#include <lib_fifo.h>
#include <lib_bufferpool.h>
#include "drv_tapi_io.h"
#include "drv_tapi_kio.h"

#if (TAPI_CFG_FEATURES & TAPI_FEAT_QOS)
   #include "drv_tapi_qos_io.h"
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_QOS) */

#if defined(LINUX) && defined (__KERNEL__)
   #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
      #include <linux/cdev.h>
   #endif /* < Linux 2.6.0 */
#endif

/* ============================= */
/* Global defines                */
/* ============================= */

/* simple usage of ptr_chk */
#define IFX_TAPI_PtrChk(ptr) ptr_chk((ptr), #ptr)

#define VMMC_IOC_MAGIC 'M'
#define VINETIC_IOC_MAGIC 'V'
#define SVIP_IOC_MAGIC 'S'
#define DUS_IOC_MAGIC 'D'
#define VXT_IOC_MAGIC 'X'

/** Maximum parameters for the hook state machine */
#define MAX_HOOKVT_PARAMS 7

#define TAPI_MIN_FLASH                80
#define TAPI_MAX_FLASH               200
#define TAPI_MIN_FLASH_MAKE          200
#define TAPI_MIN_DIGIT_LOW            30
#define TAPI_MAX_DIGIT_LOW            80
#define TAPI_MIN_DIGIT_HIGH           30
#define TAPI_MAX_DIGIT_HIGH           80
#define TAPI_MIN_OFF_HOOK             40
#define TAPI_MIN_ON_HOOK             400
#define TAPI_MIN_INTERDIGIT          300

/** time in ms while the line is in power down after the first Ground Key High */
#define GNDKH_RECHECKTIME1            10
/** time in ms to check the next Ground Key High after reenabling the line */
#define GNDKH_RECHECKTIME2            20

/** Maximum number of allowed LL drivers */
#ifndef TAPI_MAX_LL_DRIVERS
#define TAPI_MAX_LL_DRIVERS            5
#endif /* TAPI_MAX_LL_DRIVERS */

#ifndef TAPI_TONE_MAXRES
/** Maximum tone generators that can run in parallel: 3 modules ALM, PCM,
    COD and CONF and 2 tone generators */
#define TAPI_TONE_MAXRES 8
#endif

#ifndef IFX_TAPI_EVENT_POOL_INITIAL_SIZE
/** Initial number of events structures allocated by the driver. If more
    event structures are needed the pool grows automatically in steps of
    IFX_TAPI_EVENT_POOL_GROW_SIZE */
   #define IFX_TAPI_EVENT_POOL_INITIAL_SIZE 50
#endif

#ifndef IFX_TAPI_EVENT_POOL_GROW_SIZE
/** Number of events that the event structure pool grows every time it gets
    depleted. */
   #define IFX_TAPI_EVENT_POOL_GROW_SIZE 50
#endif

#ifndef IFX_TAPI_EVENT_FIFO_SIZE
/** Event Fifo Size */
   #define IFX_TAPI_EVENT_FIFO_SIZE               10
#endif /* IFX_TAPI_EVENT_FIFO_SIZE */

#ifndef IFX_TAPI_EVENT_POOL_GROW_LIMIT
   #ifdef TAPI_VERSION3
      #define IFX_TAPI_EVENT_POOL_GROW_LIMIT           500
   #else /* non TAPI_VERSION3 */
      #define IFX_TAPI_EVENT_POOL_GROW_LIMIT \
         (TAPI_MAX_LL_DRIVERS * IFX_TAPI_EVENT_FIFO_SIZE * 16/* max channels per device */)
   #endif /* TAPI_VERSION3 */
#endif

/* Number of tones for internal driver purposes */
#define TAPI_MAX_RESERVED_TONES       10
/* Maximum number of tones codes, comprising the maximum user tones and the
   maximum internal reserved tones. */
#define TAPI_MAX_TONE_CODE           (IFX_TAPI_TONE_INDEX_MAX + \
                                      TAPI_MAX_RESERVED_TONES)

/* =================================== */
/* Global typedef forward declarations */
/* =================================== */

typedef struct _TAPI_DEV      TAPI_DEV;
typedef struct _TAPI_CHANNEL  TAPI_CHANNEL;

/* ================================= */
/* Defines for tapi operating modes  */
/* ================================= */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_POLL)
/** Bit 0 used for packets and bit 1 for events. */
typedef enum _TAPI_WORKING_MODE
{
   /** Interrupt mode for packets */
   TAPI_INTERRUPT_MODE_PACKETS = 0,
   /** Interrupt mode for events */
   TAPI_INTERRUPT_MODE_EVENTS = 0,
   /** Polling mode for packets */
   TAPI_POLLING_MODE_PACKETS = 1,
   /** Polling mode for events */
   TAPI_POLLING_MODE_EVENTS = 2
} TAPI_WORKING_MODE;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_POLL) */

/* =============================== */
/* Defines for CID operations      */
/* =============================== */

/* Default indexes for CID alert tones */
enum _TAPI_CID_ALERTTONE_INDEX_DEFAULT
{
   /* Alert Tone default index, onhook */
   TAPI_CID_ATONE_INDEXDEF_ONHOOK       = IFX_TAPI_TONE_INDEX_MAX + 1,
   /* Alert Tone default index, offhook */
   TAPI_CID_ATONE_INDEXDEF_OFFHOOK      = IFX_TAPI_TONE_INDEX_MAX + 2,
   /* Alert Tone default index, AS NTT */
   TAPI_CID_ATONE_INDEXDEF_OFFHOOKNTT   = IFX_TAPI_TONE_INDEX_MAX + 3,
   /* define others cid reserved tones,
      should be in the range of reserved tones */
   TAPI_CID_ATONE_INDEXDEF_MAX          = TAPI_MAX_TONE_CODE
};

/* include debugging interface */
#include "drv_tapi_debug.h"

/* include definitions of the low level driver interface which require
   some of the definitions above */
#include "drv_tapi_ll_interface.h"

#if (TAPI_CFG_FEATURES & TAPI_FEAT_QOS)
   #include "drv_tapi_qos_ll_interface.h"
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_QOS) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_FXO)
   #include "drv_tapi_fxo_ll_interface.h"
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_FXO) */

/* Event handling */
#include "drv_tapi_event.h"

/* include for KPI requires the TAPI_CHANNEL so include it here */
#include "drv_tapi_kpi.h"

/* FXO services */
#include "drv_tapi_fxo.h"

#include "drv_tapi_ppd.h"

#define TAPI_MAX_DEVFS_HANDLES 20

typedef struct _IFX_TAPI_HL_DRV_CTX
{
   /* Pointer to the driver context provided by the LL driver. A value of
      NULL indicates that this struct is unused. */
   IFX_TAPI_DRV_CTX_t *pDrvCtx;
#if defined(LINUX) && defined (__KERNEL__)
   /* buffer to store the registered device driver name, Linux references to
      this string, e.g. on cat /proc/devices - please don't remove /Olaf */
   IFX_char_t          registeredDrvName[20];
#ifdef CONFIG_DEVFS_FS
   devfs_handle_t      TAPI_devfs_handle[TAPI_MAX_DEVFS_HANDLES];
#endif /* CONFIG_DEVFS_FS */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
   /* There is now a separate object (struct cdev) which represents char
    * devices, and which must be set up by your driver to actually make a
    * device available. */
   struct cdev *pTAPI_cdev[TAPI_MAX_DEVFS_HANDLES];
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
   struct class_device *pLL_Device[TAPI_MAX_DEVFS_HANDLES];
#else
   struct device *pLL_Device[TAPI_MAX_DEVFS_HANDLES];
#endif /* < Linux 2.6.18 */
#endif /* >= Linux 2.6.0 */
#endif /* LINUX */
} IFX_TAPI_HL_DRV_CTX_t;

/* ============================= */
/** Structure for operation control
    \internal */
/* ============================= */
typedef struct
{
   /* interrupt routine save hookstatus here */
   IFX_uint8_t               bHookState;
   /* polarity status from the line; 1 = reversed; 0 = normal */
   IFX_uint8_t               nPolarity;
   /* automatic battery switch; 1 = automatic; 0 = normal */
   IFX_uint8_t               nBatterySw;
   /* last line feed mode; is set after ringing stops */
   IFX_uint8_t               nLineMode;
   /* Last line type; changed when new line type changed with success. */
   IFX_TAPI_LINE_TYPE_t      nLineType;
   /** number of the assigned daa instance (LineType FXO),
       handled by the daa abstraction driver */
   IFX_uint16_t              nDAA;
   /** status of DAA channel initialisation
       (GPIO reservation, etc is done only once) */
   IFX_boolean_t             bDaaInitialized;
   /* set when the fault condition occurs of the device. It will be reseted
      when the line mode is modified */
   IFX_boolean_t             bFaulCond;
} TAPI_OPCONTROL_DATA_t;

/* ============================= */
/* Structure for ring data       */
/* ============================= */
struct TAPI_RING_DATA;
typedef struct TAPI_RING_DATA TAPI_RING_DATA_t;

/* ============================= */
/* Structure for pcm data        */
/* ============================= */
typedef struct
{
   /* configuration data for pcm services */
   IFX_TAPI_PCM_CFG_t   PCMConfig;
   /* save activation status */
   IFX_boolean_t        bTimeSlotActive;
   /* save configuration status */
   IFX_boolean_t        bCfgSuccess;
}TAPI_PCM_DATA_t;


/* ============================= */
/* Structure for tone data       */
/*                               */
/* ============================= */
typedef struct
{
   /* network tone on/off */
   IFX_boolean_t     bNetTone;
   /* tone playing state */
   IFX_uint32_t      nToneState;
   IFX_boolean_t     bUseSecondTG;
}TAPI_TG_TONE_DATA_t;


/* ============================= */
/* Structure for dial data       */
/*                               */
/* ============================= */
struct TAPI_DIAL_DATA;
typedef struct TAPI_DIAL_DATA TAPI_DIAL_DATA_t;

struct TAPI_GNDKH_DATA;
typedef struct TAPI_GNDKH_DATA TAPI_GNDKH_DATA_t;

/* ============================= */
/* Structure for PPD data        */
/* ============================= */
struct TAPI_PPD_DATA;
typedef struct TAPI_PPD_DATA TAPI_PPD_DATA_t;

/* ============================= */
/* Structure for metering data   */
/*                               */
/* ============================= */
struct TAPI_METER_DATA;
typedef struct TAPI_METER_DATA TAPI_METER_DATA_t;

/* ============================= */
/* Structure for internal tone   */
/* coefficients table            */
/* ============================= */
typedef enum
{
   TAPI_TONE_TYPE_NONE,
   TAPI_TONE_TYPE_SIMPLE,
   TAPI_TONE_TYPE_COMP,
   TAPI_TONE_TYPE_PREDEF,
   TAPI_TONE_TYPE_DUAL
}TAPI_TONE_TYPE;

typedef struct
{
   TAPI_TONE_TYPE          type;
   union
   {
      IFX_int32_t          nPredefined;
      IFX_TAPI_TONE_DUAL_t       dual;
      IFX_TAPI_TONE_SIMPLE_t     simple;
      IFX_TAPI_TONE_COMPOSED_t   composed;
   } tone;
}COMPLEX_TONE;

/* ============================= */
/* Structure for complex tone    */
/* data                          */
/* ============================= */
typedef struct
{
   /* type of tone sequence SIMPLE or COMPOSED */
   TAPI_TONE_TYPE          nType;
   IFX_uint32_t            nSimpleMaxReps;
   /* current repetition counter for simple tone sequence */
   IFX_uint32_t            nSimpleCurrReps;
   /* maximum number of repetitions for composed tone sequence */
   IFX_uint32_t            nComposedMaxReps;
   /* current repetition counter for composed tone sequence */
   IFX_uint32_t            nComposedCurrReps;
   /* pause time in ms betweensimple tones */
   IFX_uint32_t            nPauseTime;
   /* alternate voice timein ms between composed tones */
   IFX_uint32_t            nAlternateVoiceTime;
   /* current simple tone codeplaying */
   IFX_uint32_t            nSimpleToneCode;
   /* current composed tone code playing */
   IFX_uint32_t            nComposedToneCode;
   /* maximum simple toneswithin composed tone */
   IFX_uint32_t            nMaxToneCount;
   /* current simple tone within composed tone */
   IFX_uint32_t            nToneCounter;
   /* complex tone playing state */
   TAPI_CMPLX_TONE_STATE_t nToneState;
   /** Specifies the capability of the tone generator regarding tone
       sequence support. See \ref IFX_TAPI_TONE_RESSEQ_t for details. */
   IFX_TAPI_TONE_RESSEQ_t sequenceCap;
   /** which direction to play the tone (net or local) */
   TAPI_TONE_DST dst;
   /** Module type */
   IFX_TAPI_MODULE_TYPE_t  module;
   /** stores the index of the tone currently played */
   IFX_int32_t       nToneIndex;
   /** send IFX_TAPI_EVENT_TONE_GEN_END ? */
   IFX_boolean_t           bSendToneGenEndTapiEvent;
}TAPI_TONE_DATA_t;


/** tone resource structure as time argument */
typedef struct
{
#ifdef TAPI_HAVE_TIMERS
   /* timer id for voice path establisment */
   Timer_ID          Tone_Timer;
#endif /* TAPI_HAVE_TIMERS */
   TAPI_CHANNEL      *pTapiCh;
   /** Resource number of the tone play unit */
   IFX_uint32_t      nRes;
}TAPI_TONERES;

/* ============================== */
/* Structure for statistic data   */
/*                                */
/* ============================== */
#if (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS)
struct TAPI_STAT_DATA;
typedef struct TAPI_STAT_DATA TAPI_STAT_DATA_t;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS) */

/* ============================= */
/* channel specific structure    */
/* ============================= */
struct _TAPI_CHANNEL
{
   /* channel number */
   /* ATTENTION, nChannel must be the first element */
   IFX_uint8_t                   nChannel;
   /* pointer to the Low level driver channel */
   IFX_TAPI_LL_CH_t             *pLLChannel;
   /* pointer to the tapi device structure */
   TAPI_DEV                     *pTapiDevice;
   /* semaphore used only in blocking read access,
      in this case given from interrupt context */
   TAPI_OS_event_t               semReadBlock;
   /* wakeup queue for select on read */
   TAPI_OS_drvSelectQueue_t      wqRead;
   /* wakeup queue for select on write */
   TAPI_OS_drvSelectQueue_t      wqWrite;
   /* stores the current fax status */
   volatile IFX_boolean_t        bFaxDataRequest;

   /* PS: moved from low level channel structure */
   /* flags for different purposes, see CH_FLAGS */
   volatile IFX_uint32_t                  nFlags;

   /* channel is initialized */
   IFX_boolean_t                 bInitialized;

   /* locking semaphore for protecting data */
   TAPI_OS_mutex_t               semTapiChDataLock;

   /* overall channel protection ( read/write/ioctl level)
   PS: Avoid nested locking of this mutex. It can lead to a deadlock */
   TAPI_OS_mutex_t               semTapiChSingleIoctlAccess;

   /* data structures for services */
   TAPI_LEC_DATA_t               TapiLecAlmData;
   TAPI_LEC_DATA_t               TapiLecPcmData;
   TAPI_METER_DATA_t             *pTapiMeterData;
   TAPI_RING_DATA_t              *pTapiRingData;
   TAPI_OPCONTROL_DATA_t         TapiOpControlData;
   TAPI_PCM_DATA_t               TapiPCMData;
   TAPI_TG_TONE_DATA_t           TapiTgToneData;
   TAPI_DIAL_DATA_t              *pTapiDialData;
   TAPI_GNDKH_DATA_t             *pTapiGndkhData;
   TAPI_TONERES                  *pToneRes;
   TAPI_CID_DATA_t               *pTapiCidData;
   TAPI_PPD_DATA_t               *pTapiPpdData;

   /* complex tone data */
   TAPI_TONE_DATA_t              TapiComplexToneData [TAPI_TONE_MAXRES];
   /* Handler of event dispatcher. */
   IFX_TAPI_EVENT_HANDLER_DATA_t *pEventHandler;

   /* In Use counter */
   IFX_uint16_t                  nInUse;

#if (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET)
   FIFO_ID*                      pUpStreamFifo[IFX_TAPI_STREAM_MAX];
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_KPI)
   /** Control structure for the Kernel Packet Interface (KPI) */
   IFX_TAPI_KPI_STREAM_SWITCH    pKpiStream[IFX_TAPI_KPI_STREAM_MAX];
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_KPI) */
#ifdef EVENT_COUNTER
   /* debugging feature; per channel event sequence number */
   IFX_uint32_t nEventSeqNo;
#endif

#if (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS)
   TAPI_STAT_DATA_t              *pTapiStatData;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS) */
};

/* ============================= */
/* tapi structure                */
/* ============================= */
struct _TAPI_DEV
{
   /* channel number IFX_TAPI_DEVICE_CH_NUMBER indicates the control device */
   /* ATTENTION, nChannel must be the first element */
   IFX_uint8_t               nChannel;
   /* pointer to LL device structure */
   IFX_TAPI_LL_DEV_t        *pLLDev;
   /* link to the device driver context */
   IFX_TAPI_DRV_CTX_t       *pDevDrvCtx;

   /* Number of channels for which memory is allocated in the array below.
     This does not reflect the number of analog, signaling, pcm or coder
     channels. They are reported in the nResource struct below. */
   IFX_uint8_t               nMaxChannel;
   /* array of tapi channel structures */
   TAPI_CHANNEL             *pTapiChanelArray;
   /* struct with counters how many resources are available and nitialised */
   IFX_TAPI_RESOURCE        nResource;

   /* usage counter, counts the number of open fds */
   IFX_uint16_t              nInUse;
   /* already opened or not */
   IFX_boolean_t             bInitialized;

   /* \todo: wqEvent and bNeedWakeup should be in a separate structure */
   /* Event wakeup queue for select, this one is used to report all kind of
      device and channel events to the application. Its something like VxWorks
      version of select() call. */
   TAPI_OS_drvSelectQueue_t  wqEvent;

   /* Additional flag for vxWorks if a real wakeup is required */
   volatile IFX_boolean_t    bNeedWakeup;

   /* overall channel protection (ioctl level)
   PS: Avoid nested locking of this mutex. It can lead to a deadlock */
   TAPI_OS_mutex_t           semTapiDevSingleIoctlAccess;

#if (TAPI_CFG_FEATURES & TAPI_FEAT_POLL)
   IFX_TAPI_POLL_CONFIG_t *pPollCfg;
   /* Flag for operting mode of tapi (interrupt - default or polling). */
   TAPI_WORKING_MODE             fWorkingMode;
#if (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET)
   /* Device specific fifo for downstream direction. */
   FIFO_ID*                  pDownStreamFifo;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET) */
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_POLL)*/

   /** unique TAPI device ID [0,1,...] */
   IFX_int32_t              nDevID;

#if (TAPI_CFG_FEATURES & TAPI_FEAT_FXO)
   /* flag for SmartSLIC - required for special handling of FXO ioctls */
   IFX_boolean_t            bSmartSlicFxo;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_FXO) */

   /** last error code and error stack */
   IFX_TAPI_Error_t error;
#ifdef EVENT_COUNTER
   /* debugging feature; per device event sequence number */
   IFX_uint32_t            nEventSeqNo;
#endif
   /** Device ID */
   IFX_uint32_t             nDev;
};

/* ============================= */
/* Global variables declaration  */
/* ============================= */
/* Declarations for debug interface */
DECLARE_TRACE_GROUP  (TAPI_DRV);

/* global high level driver context used in system interface only */
extern struct _IFX_TAPI_HL_DRV_CTX gHLDrvCtx [TAPI_MAX_LL_DRIVERS];

extern const IFX_char_t TAPI_WHATVERSION[];
#ifdef HAVE_CONFIG_H
extern const IFX_char_t DRV_TAPI_WHICHCONFIG[];
#endif /* HAVE_CONFIG_H */


/* ======================================== */
/**  Driver identification                  */
/* ======================================== */
extern IFX_int32_t TAPI_Phone_Get_Version (IFX_char_t *pVer);
extern IFX_int32_t TAPI_Phone_Check_Version (IFX_TAPI_VERSION_t const *vers);

/* ======================================== */
/**  TAPI Initialisation                    */
/* ======================================== */
extern IFX_int32_t IFX_TAPI_Driver_Start (IFX_void_t);
extern IFX_void_t  IFX_TAPI_Driver_Stop (IFX_void_t);

extern IFX_int32_t IFX_TAPI_Event_On_Driver_Start (IFX_void_t);
extern IFX_void_t  IFX_TAPI_Event_On_Driver_Stop  (IFX_void_t);

extern IFX_return_t TAPI_OS_RegisterLLDrv (
                        IFX_TAPI_DRV_CTX_t *pLLDrvCtx,
                        IFX_TAPI_HL_DRV_CTX_t *pHLDrvCtx);
extern IFX_return_t TAPI_OS_UnregisterLLDrv (
                        IFX_TAPI_DRV_CTX_t *pLLDrvCtx,
                        IFX_TAPI_HL_DRV_CTX_t *pHLDrvCtx);

extern IFX_TAPI_DRV_CTX_t* IFX_TAPI_DeviceDriverContextGet (
                        IFX_int32_t Major);

extern IFX_int32_t IFX_TAPI_DeviceStart (
      TAPI_DEV *pTapiDev,
      IFX_TAPI_DEV_START_CFG_t const *pDevStartCfg
   );

extern IFX_int32_t IFX_TAPI_DeviceStop (
                        TAPI_DEV *pTapiDev);

/* ======================================== */
/**  TAPI Services                          */
/* ======================================== */

/* Driver initialisation */
extern IFX_int32_t IFX_TAPI_Phone_Init          (TAPI_DEV *pTapiDev, IFX_TAPI_CH_INIT_t const *pInit);

/* Line services */
extern IFX_int32_t TAPI_Phone_Set_LineType      (TAPI_CHANNEL *pChannel, IFX_TAPI_LINE_TYPE_CFG_t const *pCfg);
extern IFX_int32_t TAPI_Phone_Set_Linefeed      (TAPI_CHANNEL *pChannel, IFX_int32_t nMode);
extern IFX_void_t  TAPI_Phone_Linefeed_Get     (TAPI_CHANNEL *pChannel, IFX_uint8_t *nLineMode);
extern IFX_void_t TAPI_Phone_Change_Linefeed    (TAPI_CHANNEL *pChannel, IFX_int32_t nMode);
extern IFX_int32_t TAPI_LineBatteryVoltageGet   (TAPI_CHANNEL *pChannel, IFX_TAPI_LINE_BATTERY_VOLTAGE_t *pVBat);
extern IFX_int32_t TAPI_LineBatteryVoltageSet   (TAPI_CHANNEL *pChannel, IFX_int32_t nVbat);

/* Ringing Services */
extern IFX_int32_t IFX_TAPI_Ring_Initialise_Unprot (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Ring_Cleanup        (TAPI_CHANNEL *pChannel);
extern IFX_int32_t IFX_TAPI_Ring_Prepare        (TAPI_CHANNEL *pChannel);
extern IFX_int32_t IFX_TAPI_Ring_Start          (TAPI_CHANNEL *pChannel);
extern IFX_int32_t IFX_TAPI_Ring_Stop           (TAPI_CHANNEL *pChannel);
extern IFX_int32_t IFX_TAPI_Ring_DoBlocking     (TAPI_CHANNEL *pChannel);
extern IFX_int32_t IFX_TAPI_Ring_IsActive       (TAPI_CHANNEL *pChannel);
extern IFX_int32_t IFX_TAPI_Ring_SetCadence     (TAPI_CHANNEL *pChannel, IFX_uint32_t nCadence);
extern IFX_int32_t IFX_TAPI_Ring_SetCadenceHighRes (TAPI_CHANNEL *pChannel, IFX_TAPI_RING_CADENCE_t const *pCadence);
extern IFX_int32_t IFX_TAPI_Ring_SetConfig      (TAPI_CHANNEL *pChannel, IFX_TAPI_RING_CFG_t const *pRingConfig);
extern IFX_int32_t IFX_TAPI_Ring_GetConfig      (TAPI_CHANNEL *pChannel, IFX_TAPI_RING_CFG_t *pRingConfig);
extern IFX_int32_t IFX_TAPI_Ring_SetMaxRings    (TAPI_CHANNEL *pChannel, IFX_uint32_t nMaxRings);

extern IFX_int32_t IFX_TAPI_Ring_Engine_Start   (TAPI_CHANNEL *pChannel, IFX_boolean_t bStartWithInitial);
#if (TAPI_CFG_FEATURES & TAPI_FEAT_CID)
extern IFX_int32_t IFX_TAPI_Ring_CalculateRingTiming  (TAPI_CHANNEL *pChannel, IFX_uint32_t *pCadenceRingBurst, IFX_uint32_t *pCadenceRingPause);
extern IFX_void_t  IFX_TAPI_Ring_CidCadencePrepare (TAPI_CHANNEL *pChannel, IFX_TAPI_CID_SEQ_CONF_t *pCidData);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_CID) */

/* Operation Control Services */
extern IFX_int32_t TAPI_Phone_HookstateGet      (TAPI_CHANNEL *pChannel, IFX_TAPI_LINE_HOOK_STATUS_GET_t *pHookMode);
extern IFX_int32_t TAPI_Phone_LecMode_Alm_Set   (TAPI_CHANNEL *pChannel, IFX_TAPI_WLEC_CFG_t const *pLecConf);
extern IFX_int32_t TAPI_Phone_LecMode_Alm_Get   (TAPI_CHANNEL *pChannel, IFX_TAPI_WLEC_CFG_t *pWLecConf);
extern IFX_int32_t TAPI_Phone_LecMode_Pcm_Set   (TAPI_CHANNEL *pChannel, IFX_TAPI_WLEC_CFG_t const *pLecConf);
extern IFX_int32_t TAPI_Phone_LecMode_Pcm_Get   (TAPI_CHANNEL *pChannel, IFX_TAPI_WLEC_CFG_t *pLecConf);
extern IFX_int32_t TAPI_Phone_DTMFR_Cfg_Get     (TAPI_CHANNEL *pChannel, IFX_TAPI_DTMF_RX_CFG_t *pDtmfRxCoeff);
extern IFX_int32_t TAPI_Phone_DTMFR_Cfg_Set     (TAPI_CHANNEL *pChannel, IFX_TAPI_DTMF_RX_CFG_t const *pDtmfRxCoeff);
extern IFX_int32_t IFX_TAPI_PowerSave_Init      (IFX_void_t);
extern IFX_void_t  IFX_TAPI_PowerSave_Exit      (IFX_void_t);

/* Tone Services */
extern IFX_int32_t TAPI_Phone_Tone_Predef_Config (IFX_void_t);
extern IFX_int32_t IFX_TAPI_Tone_Initialise_Unprot (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Tone_Cleanup        (TAPI_CHANNEL *pChannel);
extern IFX_int32_t TAPI_Target_Tone_Play        (TAPI_CHANNEL *pChannel, IFX_TAPI_TONE_PLAY_t *pTone);
extern IFX_int32_t TAPI_Phone_Tone_Local_Play   (TAPI_CHANNEL *pChannel, IFX_int32_t nToneIndex);
extern IFX_int32_t TAPI_Phone_Tone_Local_Stop   (TAPI_CHANNEL *pChannel, IFX_int32_t nToneIndex);
extern IFX_int32_t TAPI_Phone_Tone_Net_Play     (TAPI_CHANNEL *pChannel, IFX_int32_t nToneIndex);
extern IFX_int32_t TAPI_Phone_Tone_Net_Stop     (TAPI_CHANNEL *pChannel, IFX_int32_t nToneIndex);
extern IFX_int32_t TAPI_Phone_Tone_Play         (TAPI_CHANNEL *pChannel, IFX_int32_t nToneIndex, TAPI_TONE_DST dst);
extern IFX_int32_t TAPI_Phone_Tone_Def_Stop     (TAPI_CHANNEL *pChannel, IFX_TAPI_TONE_PLAY_t *pTone);
extern IFX_int32_t TAPI_Phone_Tone_Play_Unprot  (TAPI_CHANNEL *pChannel, IFX_TAPI_TONE_PLAY_t *pTone, IFX_boolean_t bSendEndEvent);
extern IFX_int32_t TAPI_Phone_Tone_Stop         (TAPI_CHANNEL *pChannel, IFX_TAPI_TONE_PLAY_t *pTone, TAPI_TONE_DST nDirection);
#if (TAPI_CFG_FEATURES & TAPI_FEAT_DECT)
extern IFX_int32_t TAPI_DECT_Tone_Play          (TAPI_CHANNEL *pChannel, IFX_int32_t nToneIndex);
extern IFX_int32_t TAPI_DECT_Tone_Stop          (TAPI_CHANNEL *pChannel);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_DECT) */

extern IFX_int32_t TAPI_Phone_Tone_Ringback     (TAPI_CHANNEL *pChannel);
extern IFX_int32_t TAPI_Phone_Tone_Busy         (TAPI_CHANNEL *pChannel);
extern IFX_int32_t TAPI_Phone_Tone_Dial         (TAPI_CHANNEL *pChannel);

extern IFX_int32_t TAPI_Phone_Tone_Set_Level    (TAPI_CHANNEL *pChannel, IFX_TAPI_PREDEF_TONE_LEVEL_t const *pToneLevel);
extern IFX_int32_t TAPI_Phone_Tone_Get_State    (TAPI_CHANNEL *pChannel, IFX_uint32_t *pToneState);

extern IFX_int32_t TAPI_Phone_Tone_TableConf    (IFX_TAPI_TONE_t const *pTone);
extern IFX_int32_t TAPI_Phone_Add_SimpleTone    (IFX_TAPI_TONE_SIMPLE_t const *pSimpleTone);
extern IFX_int32_t TAPI_Phone_Add_ComposedTone  (IFX_TAPI_TONE_COMPOSED_t const *pComposedTone);
extern IFX_uint32_t IFX_TAPI_Tone_DurationGet   (IFX_uint32_t nToneIndex);

extern IFX_int32_t TAPI_Phone_DetectToneStart   (TAPI_CHANNEL *pChannel, IFX_TAPI_TONE_CPTD_t const *signal);
extern IFX_int32_t TAPI_Phone_DetectToneStop    (TAPI_CHANNEL *pChannel, IFX_TAPI_TONE_CPTD_t const *signal);

/* PCM Services */
extern IFX_int32_t TAPI_Phone_PCM_IF_Set_Config (TAPI_DEV *pTAPIDev, IFX_TAPI_PCM_IF_CFG_t const *pPCMif);
extern IFX_int32_t TAPI_Phone_PCM_Set_Config    (TAPI_CHANNEL *pChannel, IFX_TAPI_PCM_CFG_t const *pPCMConfig);
extern IFX_int32_t TAPI_Phone_PCM_Get_Config    (TAPI_CHANNEL *pChannel, IFX_TAPI_PCM_CFG_t *pPCMConfig);
extern IFX_int32_t TAPI_Phone_PCM_Set_Activation(TAPI_CHANNEL *pChannel, IFX_uint32_t nActive);
extern IFX_int32_t TAPI_Phone_PCM_Get_Activation(TAPI_CHANNEL *pChannel, IFX_TAPI_PCM_ACTIVATION_t *pActive);
#if (TAPI_CFG_FEATURES & TAPI_FEAT_HDLC)
extern IFX_int32_t TAPI_Phone_PCM_HDLC_Set      (TAPI_CHANNEL *pChannel, IFX_TAPI_PCM_HDLC_CFG_t const *pHdlcCfg);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_HDLC) */
extern IFX_int32_t TAPI_Phone_PCM_Loop_Set      (TAPI_CHANNEL *pChannel, IFX_TAPI_PCM_LOOP_CFG_t const *pLoopCfg);
extern IFX_int32_t TAPI_Phone_PCM_DEC_HP_Set    (TAPI_CHANNEL *pChannel, IFX_uint32_t nHp);

/* Dial Services */
extern IFX_int32_t IFX_TAPI_Dial_Initialise_Unprot (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Dial_Cleanup        (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Dial_HookEvent      (TAPI_CHANNEL *pChannel, IFX_uint8_t bHookState);
extern IFX_int32_t IFX_TAPI_Dial_SetValidationTime (TAPI_CHANNEL *pChannel, IFX_TAPI_LINE_HOOK_VT_t const *pTime);
extern IFX_int32_t IFX_TAPI_Dial_CfgApply       (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Dial_GNKH_Event     (TAPI_CHANNEL * pChannel);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_METERING)
/* Metering Services */
extern IFX_int32_t IFX_TAPI_Meter_Initialise_Unprot (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Meter_Cleanup       (TAPI_CHANNEL *pChannel);
extern IFX_boolean_t TAPI_Phone_Meter_IsActive  (TAPI_CHANNEL *pChannel);
extern IFX_int32_t TAPI_Phone_Meter_Config      (TAPI_CHANNEL *pChannel, IFX_TAPI_METER_CFG_t const *pMeterConfig);
extern IFX_int32_t TAPI_Phone_Meter_Start       (TAPI_CHANNEL *pChannel, IFX_TAPI_METER_START_t const *pMeterStart);
extern IFX_int32_t TAPI_Phone_Meter_Stop        (TAPI_CHANNEL *pChannel);
extern IFX_int32_t TAPI_Phone_Meter_Burst       (TAPI_CHANNEL *pChannel, IFX_TAPI_METER_BURST_t const *pMeterBurst);
extern IFX_int32_t TAPI_Phone_Meter_Stat        (TAPI_CHANNEL *pChannel, IFX_TAPI_METER_STATISTICS_t *pMeterStat);
extern IFX_void_t IFX_TAPI_Meter_EventServe     (TAPI_CHANNEL *pChannel);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_METERING) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_VOICE)
/* Connection services */
extern IFX_int32_t TAPI_Data_Channel_Add        (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_DATA_t const *pMap);
extern IFX_int32_t TAPI_Data_Channel_Remove     (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_DATA_t const *pMap);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_VOICE) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_VOICE)
extern IFX_int32_t TAPI_COD_ENC_CFG_Set         (TAPI_CHANNEL *pChannel, IFX_TAPI_ENC_CFG_SET_t const *pEncCfg);
extern IFX_int32_t TAPI_COD_ENC_Hold            (TAPI_CHANNEL *pChannel, IFX_int32_t nOnHold);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_VOICE) */

extern IFX_int32_t TAPI_COD_ENC_Room_Noise_Start (TAPI_CHANNEL *pChannel, IFX_TAPI_ENC_ROOM_NOISE_DETECT_t const *pDet);
extern IFX_int32_t TAPI_COD_ENC_Room_Noise_Stop (TAPI_CHANNEL *pChannel);
extern IFX_int32_t TAPI_COD_DEC_HP_Set          (TAPI_CHANNEL *pChannel, IFX_int32_t nHp);

extern IFX_int32_t IFX_TAPI_Cap_Nr_Get          (TAPI_DEV *pTapiDev, IFX_TAPI_CAP_NR_t *pCap);
extern IFX_int32_t TAPI_Last_Err_Get            (TAPI_DEV *pTapiDev, IFX_TAPI_Error_t *pErr);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_VOICE)
extern IFX_int32_t TAPI_Phone_MapAdd            (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_PHONE_t const *pMap);
extern IFX_int32_t TAPI_Phone_MapRemove         (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_PHONE_t const *pMap);
extern IFX_int32_t TAPI_Phone_PCM_MapAdd        (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_PCM_t const *pMap);
extern IFX_int32_t TAPI_Phone_PCM_MapRemove     (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_PCM_t const *pMap);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_VOICE) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_DECT)
extern IFX_int32_t TAPI_Phone_DECT_MapAdd       (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_DECT_t const *pMap);
extern IFX_int32_t TAPI_Phone_DECT_MapRemove    (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_DECT_t const *pMap);

extern IFX_int32_t IFX_TAPI_DECT_Cfg_Set        (TAPI_CHANNEL *pChannel, IFX_TAPI_DECT_CFG_t const *pDect);
extern IFX_int32_t IFX_TAPI_DECT_Enc_Cfg_Set    (TAPI_CHANNEL *pChannel, IFX_TAPI_DECT_ENC_CFG_t const *pDectEnc);
extern IFX_int32_t IFX_TAPI_DECT_Activation_Set (TAPI_CHANNEL *pChannel, IFX_uint32_t nEnable);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_DECT) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS)
extern IFX_int32_t IFX_TAPI_Stat_Initialise_Unprot (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Stat_Cleanup        (TAPI_CHANNEL *pChannel);
extern IFX_void_t  IFX_TAPI_Stat_Reset          (TAPI_CHANNEL *pChannel);
extern IFX_uint32_t IFX_TAPI_Stat_Get           (TAPI_CHANNEL *pChannel, IFX_TAPI_STREAM_t stream, TAPI_STAT_COUNTER_t counter);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS) */

extern IFX_int32_t IFX_TAPI_Module_Find_Connected_Data_Channel (TAPI_CHANNEL *pChannel, IFX_TAPI_MAP_TYPE_t nModType, TAPI_CHANNEL **pTapiCh);

/* Event packet generation */
extern IFX_int32_t TAPI_EVENT_PKT_EV_Generate   (TAPI_CHANNEL *pChannel, IFX_TAPI_PKT_EV_GENERATE_t const *pPacketEvent);

extern IFX_int32_t TAPI_EVENT_PKT_EV_Oob_Dtmf_Set (TAPI_CHANNEL *pChannel, IFX_uint32_t nOobMode);
extern IFX_int32_t TAPI_EVENT_PKT_EV_Oob_Mftd_Set (TAPI_CHANNEL *pChannel, IFX_TAPI_PKT_EV_OOB_MFTD_t const *pOob);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET)
extern IFX_int32_t TAPI_PKT_Flush               (TAPI_CHANNEL* pChannel);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET) */

extern IFX_int32_t TAPI_Test_Hook_Gen           (TAPI_CHANNEL *pChannel, IFX_uint32_t nHook);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38)
extern IFX_void_t TAPI_FaxT38_Event_Update      (TAPI_CHANNEL *pChannel, IFX_uint8_t status, IFX_uint8_t error);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38) */

/* FXS Phone Plug Detection */
#if (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION)
extern IFX_int32_t IFX_TAPI_PPD_Cfg_Get        (TAPI_CHANNEL *pChannel, IFX_TAPI_LINE_PHONE_DETECT_CFG_t *pPpdConf);
extern IFX_int32_t IFX_TAPI_PPD_Cfg_Set        (TAPI_CHANNEL *pChannel, IFX_TAPI_LINE_PHONE_DETECT_CFG_t const *pPpdConf);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION) */

/* ======================================== */
/**  Other functions                        */
/* ======================================== */
extern IFX_boolean_t ptr_chk(const IFX_void_t const *ptr, const IFX_char_t *pPtrName);

extern IFX_void_t TAPI_Tone_Step_Completed(TAPI_CHANNEL* pChannel, IFX_uint8_t utgNum);
extern IFX_int32_t TAPI_DeferWork (IFX_void_t* pFunc, IFX_void_t* pParam);

#endif  /* DRV_TAPI_H */
