/******************************************************************************

                              Copyright (c) 2009
                            Lantiq Deutschland GmbH
                     Am Campeon 3; 85579 Neubiberg, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file drv_tapi_ioctl.c
   This file contains the implementation of the TAPI ioctl handling.
*/
/* ============================= */
/* Includes                      */
/* ============================= */

#ifdef EVENT_LOGGER_DEBUG
   #ifdef LINUX
      #include <asm/ioctl.h>
   #endif
   #ifdef VXWORKS
      #include "ioctl.h"
   #endif
#endif /* EVENT_LOGGER_DEBUG */

#if (defined(IFXOS_USE_DEV_IO) && (IFXOS_USE_DEV_IO == 1))
   #include <ifxos_device_access.h>
   #ifdef LINUX
       #include <asm/ioctl.h> /** \todo Fix it */
   #endif
#endif

#include "drv_tapi.h"
#include "drv_tapi_errno.h"

#include "drv_tapi_polling.h"
#include "drv_tapi_cid.h"
#include "drv_tapi_qos.h"
#include "drv_tapi_ppd.h"

#include "drv_tapi_ioctl.h"
#define TAPI_DECLARE_IOCTL_HANDLERS_MAP
#include "drv_tapi_ioctl_handlers.h"

/* ============================= */
/* Local Macros & Definitions    */
/* ============================= */

/* ============================= */
/* Local type definition         */
/* ============================= */

/** states of IOCTL loging */
typedef enum
{
   /** before ioctl handling */
   TAPI_IOCTL_PRELOG,
   /** after ioctl handling */
   TAPI_IOCTL_POSTLOG
} TAPI_IOCTL_LOG_t;

#ifdef TAPI_ONE_DEVNODE
/** Generic structure for channel based ioctl arguments. */
typedef struct
{
   /** Device index */
   IFX_uint16_t dev;
   /** Channel "module" index */
   IFX_uint16_t ch;
   /** Any parameter used by ioctls */
   IFX_uint32_t param;
} TAPI_CH_IO_ARG_t;

/** Generic structure for device based ioctl arguments. */
typedef struct
{
   /** Device index */
   IFX_uint16_t dev;
   /** Any parameter used by ioctls */
   IFX_uint32_t param;
} TAPI_DEV_IO_ARG_t;
#endif /* TAPI_ONE_DEVNODE */

/* ============================= */
/* Global variable definition    */
/* ============================= */

/* ============================= */
/* Local function declaration    */
/* ============================= */

static IFX_int32_t tapi_Dev_Spec_Ioctl (IFX_TAPI_ioctlCtx_t *pCtx);

static IFX_int32_t tapi_Spec_Ioctl (IFX_TAPI_ioctlCtx_t *pCtx);

/* ============================= */
/* Local variable definition     */
/* ============================= */

/* ============================= */
/* Local function definition     */
/* ============================= */

#ifdef EVENT_LOGGER_DEBUG
static IFX_void_t tapi_IoctlLog (
   IFX_TAPI_ioctlCtx_t *pCtx,
   IFX_int32_t err,
   TAPI_IOCTL_LOG_t mode)
{
   IFX_uint16_t nDev = 0, nCh = 0;

   nCh = pCtx->pTapiCh->nChannel;
   nDev = pCtx->pTapiCh->pTapiDevice->nDev;

   switch (mode)
   {
      case TAPI_IOCTL_PRELOG:
         /* log read ioctl before it is executed, so that we see it in case
          * of failure */
         if (IFX_TRUE == pCtx->bRead)
         {
            LOG_RD_IOCTL(nDev, nCh, pCtx->nCmd, IFX_NULL, 0, err);
         }
         else if (IFX_TRUE == pCtx->bWrite)
         {
            IFX_void_t *pArg = IFX_NULL;

            if (pCtx->nArgSize > sizeof (IFX_uint32_t))
            {
               /* we are sure that passed structure */
               pArg = (IFX_void_t *)pCtx->nArg;
            }
            else if (pCtx->nArgSize > 0)
            {
               /* simple integer in argument,
                  or small structure (should not be) */
               pArg = (IFX_void_t *)&pCtx->nArg;
            }

            LOG_WR_IOCTL(nDev, nCh, pCtx->nCmd, pArg, pCtx->nArgSize, err);
         }
         else
         {
            LOG_WR_IOCTL (nDev, nCh, pCtx->nCmd, IFX_NULL, 0, err);
         }

         break;
      case TAPI_IOCTL_POSTLOG:
         if (IFX_FALSE == pCtx->bRead)
            break;

         /* \note: IFX_TAPI_CAP_CHECK - returns 1 if capability is supported
            IFX_TAPI_RING - returns 1 if phone hooked off
          */
         if ((err == 1) &&
             ((IFX_TAPI_CAP_CHECK == pCtx->nCmd) ||
              (IFX_TAPI_RING == pCtx->nCmd)))
         {
            err = IFX_SUCCESS;
         }

         LOG_RD_IOCTL(nDev, nCh, pCtx->nCmd, pCtx->nArg, pCtx->nArgSize, err);
         break;
   }
}
#endif /* EVENT_LOGGER_DEBUG */

/* ============================= */
/* Global function definition    */
/* ============================= */

/**
   TAPI global IOCTL handling

   \param  pCtx         Pointer to IOCTL context.

   \return
   - TAPI_statusOk if successful
   - TAPI_statusErr in case of error
   - TAPI_statusCtxErr
   - TAPI_statusParam

   \remarks
   This function does the following functions:
      - If the ioctl command is device specific, low-level driver's ioctl function
      - If the ioctl command is TAPI specific, it is handled at this level
*/
IFX_int32_t  TAPI_Ioctl (IFX_TAPI_ioctlCtx_t *pCtx)
{
   IFX_int32_t ret = TAPI_statusOk;

   TAPI_ASSERT (pCtx);

   if (IFX_TAPI_PtrChk (pCtx->pIOCtl))
   {
#ifdef EVENT_LOGGER_DEBUG
      /* \note: logging of TAPI known ioctls only */
      tapi_IoctlLog (pCtx, TAPI_statusOk, TAPI_IOCTL_PRELOG);
#endif /* EVENT_LOGGER_DEBUG */

      /* TAPI specific ioctl */
      ret = tapi_Spec_Ioctl (pCtx);

#ifdef EVENT_LOGGER_DEBUG
      tapi_IoctlLog (pCtx, ret, TAPI_IOCTL_POSTLOG);
#endif /* EVENT_LOGGER_DEBUG */

      /* \note: IFX_TAPI_CAP_CHECK - returns 1 if capability is supported
         IFX_TAPI_RING - returns 1 if phone hooked off
       */
      if ((IFX_TAPI_CAP_CHECK == pCtx->nCmd) ||
          (IFX_TAPI_RING == pCtx->nCmd))
         return ret;
   }
   else
   {
      /* \note: logging of TAPI unknown ioctls should be done by ll-driver */
      /* Low-level device specific ioctl */
      ret = tapi_Dev_Spec_Ioctl (pCtx);
   }

   if (!TAPI_SUCCESS (ret))
      return IFX_ERROR;

   return IFX_SUCCESS;
}

/**
   Retrieve channel and device number by IOCTL conext

   \param   pDrvCtx     Handle to the driver context
   \param   pCtx        Handle to the IOCTL context
   \param   pnDevice    Handle to the result of device number
                        For real device used range [-1..n)
                        -1 value mean no device detected
   \param   pnChannel   Handle to the result of channel number
                        For real channel used range [-1..n)
                        -1 value mean no channel detected

   \return
      IFX_SUCCESS on success

   \remarks
      Single device node find out the device and channel from the
      given structure only for known IOCTL. Otherwise used minor number.
      Following is supported:
      - (1) driver ioctl using simple integers or structures
      - (2) device ioctl using structures with 'dev' fields
      - (3) channel ioctl using structures with 'dev' and 'ch' fields
*/
static IFX_return_t tapi_ioctlDevChGet (
   IFX_TAPI_DRV_CTX_t *pDrvCtx,
   IFX_TAPI_ioctlCtx_t *pCtx,
   IFX_int32_t *pnDevice,
   IFX_int32_t *pnChannel)
{
   TAPI_ASSERT (pDrvCtx);
   TAPI_ASSERT (pCtx);
   TAPI_ASSERT (pnDevice);
   TAPI_ASSERT (pnChannel);

   /* set default value */
   *pnDevice = -1;
   *pnChannel = -1;

#ifndef TAPI_ONE_DEVNODE

   /* retrieve device and channel number by device node */
   *pnDevice = (pCtx->nMinor / pDrvCtx->minorBase) - 1;
   *pnChannel = (pCtx->nMinor % pDrvCtx->minorBase) - 1;

   if (IFX_TAPI_PtrChk (pCtx->pIOCtl))
   {
      /* correct context according to the known IOCTL handler */
      if ((TAPI_IOC_CTX_NONE == pCtx->pIOCtl->nContext) ||
          (TAPI_IOC_CTX_DRV == pCtx->pIOCtl->nContext))
      {
         *pnDevice = -1;
      }
      else if (TAPI_IOC_CTX_DEV == pCtx->pIOCtl->nContext)
      {
         *pnChannel = -1;
      }
   }

#else /* TAPI_ONE_DEVNODE */

   if (IFX_TAPI_PtrChk (pCtx->pIOCtl))
   {
      TAPI_CH_IO_ARG_t *pArg = (TAPI_CH_IO_ARG_t*)pCtx->nArg;

      if (TAPI_IOC_CTX_NONE == pCtx->pIOCtl->nContext)
      {
         /* context independent IOCTLs do not contain 'dev' and 'ch' fields in argument */
         return IFX_SUCCESS;
      }

      if (TAPI_IOC_CTX_DRV == pCtx->pIOCtl->nContext)
      {
         /* driver specific IOCTLs do not contain 'dev' and 'ch' fields in argument */
         return IFX_SUCCESS;
      }

      /*  retrieve correct device for channel and device specific channel,
          argument contain 'dev' field */
      if (pArg->dev == IFX_TAPI_EVENT_ALL_DEVICES)
      {
         /* use first device */
         *pnDevice = 0;
      }
      else
      {
         *pnDevice = pArg->dev;
      }

      if (TAPI_IOC_CTX_DEV == pCtx->pIOCtl->nContext)
      {
         /* device specific IOCTLs do not contain 'ch' field in argument */
         return IFX_SUCCESS;
      }

      /*  retrieve correct channel for channel specific IOCTL,
          argument contain 'ch' field */
      if (pArg->ch == IFX_TAPI_EVENT_ALL_CHANNELS)
      {
         /* use firs channel */
         *pnChannel = 0;
      }
      else
      {
         *pnChannel = pArg->ch;
      }
   }
#endif /* TAPI_ONE_DEVNODE */

   return IFX_SUCCESS;
}

/*
   Retrieve IOCTL argument from user

   \param pDrvCtx    Handle for driver context
   \param ioarg      IOCTL argument
   \param pCtx       Handle to the IOCTL context to be updated

   \return
      IFX_SUCCESS on success
 */
static IFX_return_t tapi_ioctlArgGet (
   IFX_TAPI_DRV_CTX_t *pDrvCtx,
   IFX_ulong_t ioarg,
   IFX_TAPI_ioctlCtx_t *pCtx)
{
   TAPI_ASSERT (pCtx);

   /* store argument in the context */
   pCtx->nArg = ioarg;

   if (!IFX_TAPI_PtrChk (pCtx->pIOCtl))
   {
      /* TAPI high-level do not know anything about low-level services,
         just pass it through */
      return IFX_SUCCESS;
   }

   /*
    * Prepare argument for known service
    */

   /* Retrieve IOCTL argument size from IOCTL declaration */
   pCtx->nArgSize = TAPI_IOC_SIZE (pCtx->nCmd);

#if !defined (TAPI_ONE_DEVNODE)
   if ((IFX_TAPI_TONE_STOP == pCtx->nCmd) && (ioarg < 255))
   {
      /* \note: That hack required to support integer arguments
         for 'IFX_TAPI_TONE_STOP' service, the pointer also supported */
      IFX_TAPI_TONE_PLAY_t tmpTone;

      /* prepare argument */
      memset (&tmpTone, 0, sizeof (tmpTone));
      tmpTone.index = (IFX_uint32_t)ioarg;
      tmpTone.module = IFX_TAPI_MODULE_TYPE_COD;

      /* replace real argument by temporary representation */
      ioarg = (IFX_ulong_t)&tmpTone;

      /* correct argument size */
      pCtx->nArgSize = sizeof (tmpTone);
   }
   else if (IFX_TAPI_TONE_CPTD_STOP == pCtx->nCmd)
   {
      /* \note: That hack required to support ONLY NONE arguments
         for 'IFX_TAPI_TONE_CPTD_STOP' service */
      IFX_TAPI_TONE_CPTD_t tmpCptd;

      /* prepare argument */
      memset (&tmpCptd, 0, sizeof (tmpCptd));
      tmpCptd.tone = 1; /* dummy tone, have no sense */

      /* replace real argument by temporary representation */
      ioarg = (IFX_ulong_t)&tmpCptd;

      /* correct argument size */
      pCtx->nArgSize = sizeof (tmpCptd);
   }
   else if (IFX_TAPI_CID_RX_START == pCtx->nCmd)
   {
      /* \note: That hack required to support integer arguments
         for 'IFX_TAPI_CID_RX_START' service */
      IFX_TAPI_CID_RX_CFG_t tmpCid;

      /* prepare argument */
      memset (&tmpCid, 0, sizeof (tmpCid));
      tmpCid.hookMode = (IFX_TAPI_CID_HOOK_MODE_t)ioarg;
      tmpCid.module = IFX_TAPI_MODULE_TYPE_ALM;

      /* replace real argument by temporary representation */
      ioarg = (IFX_ulong_t)&tmpCid;

      /* correct argument size */
      pCtx->nArgSize = sizeof (tmpCid);
   }
   else if (IFX_TAPI_CID_RX_STOP == pCtx->nCmd)
   {
      /* \note: That hack required to support ONLY NONE arguments
         for 'IFX_TAPI_CID_RX_STOP' service */
      IFX_TAPI_CID_RX_CFG_t tmpCid;

      /* prepare argument */
      memset (&tmpCid, 0, sizeof (tmpCid));
      /* no special initialization required, have no sense */

      /* replace real argument by temporary representation */
      ioarg = (IFX_ulong_t)&tmpCid;

      /* correct argument size */
      pCtx->nArgSize = sizeof (tmpCid);
   }
#endif /* !defined (TAPI_ONE_DEVNODE) */

#if defined (USE_OLD_TAPI_CAP_LIST)
   if (IFX_TAPI_CAP_LIST == pCtx->nCmd)
   {
      IFX_TAPI_CAP_LIST_t tmpList;
      /* \note: temporary support old style
         for 'IFX_TAPI_CAP_LIST' service */

      /* prepare argument */
      memset (&tmpList, 0, sizeof (tmpList));

      tmpList.nCap = pDrvCtx->CAP_Number_Get (pDrvCtx->pTapiDev[0].pLLDev);
      tmpList.pList = (IFX_TAPI_CAP_t*)ioarg;

      /* replace real argument by temporary representation */
      ioarg = (IFX_ulong_t)&tmpList;

      /* correct argument size */
      pCtx->nArgSize = sizeof (tmpList);
   }
#endif /* !defined (TAPI_ONE_DEVNODE) */

   if (TAPI_ARG_POINTER == pCtx->nArgKind)
   {
      /* check for integer value in IOCTL argument
         \note:
            not working for IFX_TAPI_RING_CADENCE_SET,
            because cadence can looks like address. */
      if ((pCtx->nArgSize <= sizeof (IFX_uint32_t)) &&
          (ioarg > 0) && (ioarg <= 255))
      {
         /* integer value can not be used as pointer */
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Integer argument detected for ioctl %s, "
               "required pointer." TAPI_CRLF,
               TAPI_ioctlNameGet (pCtx->nCmd)));
         return IFX_ERROR;
      }

      /* check for NULL pointer in IOCTL argument */
      if (0 == ioarg)
      {
         /* integer value can not be used as pointer */
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "NULL pointer detected for ioctl %s" TAPI_CRLF,
               TAPI_ioctlNameGet (pCtx->nCmd)));
         return IFX_ERROR;
      }
   }
   else
   {
      /* non pointer values do not require copy from/to user-space */
      pCtx->bArgUsrCpy = IFX_FALSE;
   }

   /* Do direct copy in to the local memory
      if ioctl argument was locally changed or
      copy from user-space required */
   if ((ioarg != pCtx->nArg) || (IFX_TRUE == pCtx->bArgUsrCpy))
   {
      IFX_void_t *pArg;

      TAPI_ASSERT (pCtx->nArgSize);

      pArg = TAPI_OS_Malloc (pCtx->nArgSize);
      if (!IFX_TAPI_PtrChk (pArg))
      {
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Cannot allocate memory for ioctl %s." TAPI_CRLF,
               TAPI_ioctlNameGet (pCtx->nCmd)));
         return IFX_ERROR;
      }

      /* copy ioctl argument in to the local memory */
      if (ioarg != pCtx->nArg)
      {
         memcpy (pArg, (IFX_void_t*)ioarg, pCtx->nArgSize);
      }
      else if (!TAPI_OS_CpyUsr2Kern (pArg, (IFX_void_t*)ioarg, pCtx->nArgSize))
      {
         TAPI_OS_Free (pArg);
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Cannot copy argument for ioctl %s." TAPI_CRLF,
               TAPI_ioctlNameGet (pCtx->nCmd)));
         return IFX_ERROR;
      }

      pCtx->nArg = (IFX_ulong_t)pArg;
   }

#if !defined (TAPI_ONE_DEVNODE)
   if (IFX_TAPI_CH_INIT == pCtx->nCmd)
   {
      /* \note: That hack required to correct the 'ch' field
         for 'IFX_TAPI_CH_INIT' service backward compatibility */
      ((IFX_TAPI_CH_INIT_t*)pCtx->nArg)->ch = \
         (pCtx->nMinor % pDrvCtx->minorBase) - 1;
   }
#endif /* !defined (TAPI_ONE_DEVNODE) */

   return IFX_SUCCESS;
}

/*
   Put IOCTL argument back in to the user

   \param ioarg      IOCTL argument
   \param pCtx       Handle to the IOCTL context to be updated
 */
static IFX_void_t tapi_ioctlArgPut (
   IFX_ulong_t ioarg,
   IFX_TAPI_ioctlCtx_t *pCtx)
{
   TAPI_ASSERT (pCtx);

   /* do nothing if direct argument was used */
   if (ioarg == pCtx->nArg)
      return;

   do {
#if defined (USE_OLD_TAPI_CAP_LIST)
      if (IFX_TAPI_CAP_LIST == pCtx->nCmd)
         break;
#endif /* !defined (USE_OLD_TAPI_CAP_LIST) */

      /* only read direction can require copy result back */
      if (IFX_FALSE == pCtx->bRead)
         break;

      /* copy result back to the user */
      if (IFX_FALSE == pCtx->bArgUsrCpy)
      {
         memcpy ((IFX_void_t*)ioarg, (IFX_void_t*)pCtx->nArg,
            pCtx->nArgSize);
      }
      else
      {
         TAPI_OS_CpyKern2Usr ((IFX_void_t*)ioarg, (IFX_void_t*)pCtx->nArg,
            pCtx->nArgSize);
      }
   } while (0);

   /* release local memory */
   TAPI_OS_Free ((IFX_void_t*)pCtx->nArg);
}

/**
   Prepare IOCTL context

   \param pDrvCtx    Handle for driver context
   \param nMinor     Minor number of device node
   \param iocmd      IOCTL command
   \param ioarg      IOCTL argument
   \param bArgUsrCpy IOCTL argument require copy from/to user space
   \param pCtx       Handle to the IOCTL context to be updated

   \return
      IFX_SUCCESS on success
*/
IFX_return_t TAPI_ioctlContextGet (
   IFX_TAPI_DRV_CTX_t *pDrvCtx,
   IFX_uint32_t nMinor,
   IFX_uint32_t iocmd,
   IFX_ulong_t ioarg,
   IFX_boolean_t bArgUsrCpy,
   IFX_TAPI_ioctlCtx_t *pCtx)
{
   IFX_int32_t nDevice = -1;
   IFX_int32_t nChannel = -1;

   TAPI_ASSERT (pDrvCtx);
   TAPI_ASSERT (pCtx);

   /* do basic initialization */
   memset (pCtx, 0, sizeof(IFX_TAPI_ioctlCtx_t));

   /* store TAPI service identifier */
   pCtx->nCmd = iocmd;
   /* remember storage space of argument value */
   pCtx->bArgUsrCpy = bArgUsrCpy;
#if !defined (TAPI_ONE_DEVNODE)
   /* store minor number for future intermediate use */
   pCtx->nMinor = nMinor;
#endif /* !defined (TAPI_ONE_DEVNODE) */

   /* retrieve IOCTL descriptor */
   switch (TAPI_IOC_MAGIC (iocmd))
   {
#if (TAPI_CFG_FEATURES & TAPI_FEAT_QOS)
      case QOS_IOC_MAGIC:
         /* This group contains extra TAPI IOCTLs that are handled just like
            all other TAPI IOCTLs below. */

         if (TAPI_IOC_IDX (iocmd) >= ARRAY_SIZE (gQosIOCtls))
         {
            TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Unknown QOS IOCTL command %s" TAPI_CRLF,
               TAPI_ioctlNameGet (iocmd)));
            return IFX_ERROR;
         }

         /* retrieve corresponding IOCTL descriptor */
         pCtx->pIOCtl = gQosIOCtls + TAPI_IOC_IDX (iocmd);
         break;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_QOS) */
      case IFX_TAPI_IOC_MAGIC:
         if (TAPI_IOC_IDX (iocmd) >= ARRAY_SIZE (gIOCtls))
         {
            TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Unknown IOCTL command %s" TAPI_CRLF,
                TAPI_ioctlNameGet (iocmd)));
            return IFX_ERROR;
         }

         /* retrieve corresponding IOCTL descriptor */
         pCtx->pIOCtl = gIOCtls + TAPI_IOC_IDX (iocmd);
         break;
      default:
         /* NOTE: TAPI do not know the argument structure of low-level IOCTLs,
            just forward to the low-level handler without device or channel detection.
            low-level driver responsible for it own target detection */
         break;
   }

   /* prepare context for known IOCTL */
   if (IFX_TAPI_PtrChk (pCtx->pIOCtl))
   {
#ifdef DEBUG
      TRACE(TAPI_DRV, DBG_LEVEL_LOW, ("%s -> %s [%d]" TAPI_CRLF,
            TAPI_ioctlNameGet (iocmd), pCtx->pIOCtl->info, pCtx->pIOCtl->line));
#endif /* DEBUG */
      /* detect IOCTL read direction */
      pCtx->bRead = TAPI_IOC_READ (iocmd);
      /* detect IOCTL write direction */
      pCtx->bWrite = TAPI_IOC_WRITE (iocmd);

      /* for integer argument write direction available only */
      if ((TAPI_ARG_INTEGER == pCtx->pIOCtl->nArgKind) &&
          (IFX_TRUE == pCtx->bRead))
      {
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Incorrect direction detected for ioctl %s." TAPI_CRLF,
               TAPI_ioctlNameGet (iocmd)));
         return IFX_ERROR;
      }

      /* detect the expected argument type */
      pCtx->nArgKind = pCtx->pIOCtl->nArgKind;

#if defined (TAPI_ONE_DEVNODE)
      if ((TAPI_IOC_CTX_DEV == pCtx->pIOCtl->nContext) ||
          (TAPI_IOC_CTX_CH == pCtx->pIOCtl->nContext))
      {
         /* Device and channel specific service for single device node
            always using pointer as argument. */
         pCtx->nArgKind = TAPI_ARG_POINTER;
      }
#endif /* TAPI_ONE_DEVNODE */
   }

   /* retrieve argument for TAPI IOCTL */
   if (IFX_SUCCESS != tapi_ioctlArgGet (pDrvCtx, ioarg, pCtx))
   {
      return IFX_ERROR;
   }

   /* retrieve destination for TAPI IOCTL */
   if (IFX_SUCCESS != tapi_ioctlDevChGet (pDrvCtx, pCtx, &nDevice, &nChannel))
   {
      return IFX_ERROR;
   }

   if (nDevice >= pDrvCtx->maxDevs)
   {
      TRACE(TAPI_DRV, DBG_LEVEL_NORMAL, ("TAPI ERROR: "
         "Unsupported device '%d' used for '%s' ioctl %s" TAPI_CRLF,
         nDevice, pDrvCtx->drvName, TAPI_ioctlNameGet (iocmd)));
      return IFX_ERROR;
   }

   if (nChannel >= pDrvCtx->maxChannels)
   {
      TRACE(TAPI_DRV, DBG_LEVEL_NORMAL, ("TAPI ERROR: "
         "Unsupported channel '%d' used for '%s' ioctl %s" TAPI_CRLF,
         nChannel, pDrvCtx->drvName, TAPI_ioctlNameGet (iocmd)));
      return IFX_ERROR;
   }

   /* assign handle to the IOCTL context */
   if (nDevice < 0)
   {
      if (IFX_TAPI_PtrChk (pCtx->pIOCtl) &&
          (TAPI_IOC_CTX_NONE == pCtx->pIOCtl->nContext))
      {
         /* context independent */
         pCtx->nContext = TAPI_IOC_CTX_NONE;
      }
      else
      {
         /* driver context */
         pCtx->nContext = TAPI_IOC_CTX_DRV;
      }

      /* took first channel of first device */
      nDevice = 0;
      nChannel = 0;
   }
   else if (nChannel < 0)
   {
      /* device context */
      pCtx->nContext = TAPI_IOC_CTX_DEV;
      /* took first channel */
      nChannel = 0;
   }
   else
   {
      /* channel context */
      pCtx->nContext = TAPI_IOC_CTX_CH;
   }

   /* validate declared and detected service description */
   if (IFX_TAPI_PtrChk (pCtx->pIOCtl))
   {
      /*  check the detected and available contexts */
      if (pCtx->nContext != pCtx->pIOCtl->nContext)
      {
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Incorrect context detected [%d], "
               "expected [%d] for ioctl %s." TAPI_CRLF,
               pCtx->nContext, pCtx->pIOCtl->nContext,
               TAPI_ioctlNameGet (iocmd)));
         /* Service called with wrong context fd */
         return IFX_ERROR;
      }
   }

   /* retrieve context handle */
   pCtx->pTapiCh = pDrvCtx->pTapiDev[nDevice].pTapiChanelArray + nChannel;

   return IFX_SUCCESS;
}

/**
   Finalize IOCTL context communication

   \param ioarg      IOCTL argument
   \param pCtx       Handle to the IOCTL context to be updated
*/
IFX_void_t TAPI_ioctlContextPut (
   IFX_ulong_t ioarg,
   IFX_TAPI_ioctlCtx_t *pCtx)
{
   /* handle argument value */
   tapi_ioctlArgPut (ioarg, pCtx);

   /* do cleanup memory */
   memset (pCtx, 0, sizeof(IFX_TAPI_ioctlCtx_t));
}

/**
   Device Specific IOCTL handling

   \param  pCtx         Pointer to IOCTL context.

   \return
   - TAPI_statusOk if successful
   - TAPI_statusErr in case of error
*/
static IFX_int32_t tapi_Dev_Spec_Ioctl (IFX_TAPI_ioctlCtx_t* pCtx)
{
   TAPI_DEV       *pTapiDev = IFX_NULL;
   IFX_void_t     *pLLHnd = IFX_NULL;
   IFX_int32_t    retLL = TAPI_statusOk;

   TAPI_ASSERT (pCtx);
   TAPI_ASSERT (pCtx->pTapiCh);

   /* retrieve current device handle */
   pTapiDev = pCtx->pTapiCh->pTapiDevice;

#ifdef TAPI_ONE_DEVNODE
   /* \note: low level handler should manually detect context,
      pass IFX_NULL as ioctl handle.
      */
#else
   /* Provide either a low-level device or channel context to the low-level
      IOCTL handler. The low level IOCTL handler has to check the correctness
      of the context. */
   switch (pCtx->nContext)
   {
      case TAPI_IOC_CTX_CH:
         pLLHnd = (IFX_void_t*) pCtx->pTapiCh->pLLChannel;
         break;
      case TAPI_IOC_CTX_DRV:
         /* use device handle for driver context */
         /*lint -fallthrough*/
      case TAPI_IOC_CTX_NONE:
         /* use device handle for context independent service */
         /*lint -fallthrough*/
      case TAPI_IOC_CTX_DEV:
         pLLHnd = (IFX_void_t*) pTapiDev->pLLDev;
         break;
      default:
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Unsupported context detected [%d] for ll-ioctl %s" TAPI_CRLF,
               pCtx->nContext, TAPI_ioctlNameGet (pCtx->nCmd)));
         /* should never occur */
         return TAPI_statusCtxErr;
   }
#endif /* not TAPI_ONE_DEVNODE */

   /* make sure that the low-level supports IOCTLs */
   if (!IFX_TAPI_PtrChk (pTapiDev->pDevDrvCtx->Ioctl))
      RETURN_DEVSTATUS (TAPI_statusNoIoctl, 0);

   /* dispatch to low level IOCTL handler */
   retLL = pTapiDev->pDevDrvCtx->Ioctl (pLLHnd, pCtx->nCmd, pCtx->nArg);

   if (!TAPI_SUCCESS(retLL))
   {
      /* if low-level failed add a generic high-level error code */
      RETURN_DEVSTATUS (TAPI_statusLLFailed, retLL);
   }

   return TAPI_statusOk;
}

/**
   TAPI specific IOCTL handling

   \param  pCtx         Pointer to IOCTL context.

   \return
   - TAPI_statusOk if successful
   - TAPI_statusErr in case of error
   - TAPI_statusCtxErr
   - TAPI_statusParam

   \remarks
      Only TAPI IOCTLs with appropriate TAPI magic number are
      handled within this function.
*/
static IFX_int32_t  tapi_Spec_Ioctl (IFX_TAPI_ioctlCtx_t *pCtx)
{
   IFX_int32_t    ret   = TAPI_statusOk, retLL = TAPI_statusOk;
   TAPI_DEV       *pTapiDev = IFX_NULL;
   IFX_void_t     *pHnd = IFX_NULL,
                  *pLLHnd = IFX_NULL,
                  *pArg = IFX_NULL;
   TAPI_Handler_t pHandler;
   TAPI_OS_mutex_t *pSemSingleIoctlAccess = IFX_NULL;

   /* retrieve current device handle */
   pTapiDev = pCtx->pTapiCh->pTapiDevice;
   pArg = (IFX_void_t*)pCtx->nArg;

   /* retrieve device handle for error reports */
   switch (pCtx->nContext)
   {
      case TAPI_IOC_CTX_CH:
         pHnd = (IFX_void_t *) pCtx->pTapiCh;
         pLLHnd = (IFX_void_t *) pCtx->pTapiCh->pLLChannel;
         pSemSingleIoctlAccess = &pCtx->pTapiCh->semTapiChSingleIoctlAccess;
         break;
      case TAPI_IOC_CTX_DEV:
         pHnd = (IFX_void_t *) pTapiDev;
         pLLHnd = (IFX_void_t*) pTapiDev->pLLDev;
         pSemSingleIoctlAccess = &pTapiDev->semTapiDevSingleIoctlAccess;
         break;
      case TAPI_IOC_CTX_DRV:
         pHnd = (IFX_void_t *) pTapiDev->pDevDrvCtx;
         pLLHnd = IFX_NULL;
         pSemSingleIoctlAccess = IFX_NULL;
         break;
      case TAPI_IOC_CTX_NONE:
         pHnd = IFX_NULL;
         pLLHnd = IFX_NULL;
         pSemSingleIoctlAccess = IFX_NULL;
         break;
      default:
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH, ("TAPI ERROR: "
               "Unsupported context detected [%d] for hl-ioctl %s" TAPI_CRLF,
               pCtx->nContext,
               TAPI_ioctlNameGet (pCtx->nCmd)));
         /* should never occur */
         return TAPI_statusCtxErr;
   }

   if (IFX_FALSE == pTapiDev->bInitialized)
   {
      /* check requirement for initialized device */
      if (TAPI_DS_INIT == (TAPI_DS_INIT & pCtx->pIOCtl->nDevState))
      {
         TRACE(TAPI_DRV, DBG_LEVEL_LOW,
               ("IOCTL %s for dev%d blocked." TAPI_CRLF,
               TAPI_ioctlNameGet (pCtx->nCmd), pTapiDev->nDev));
         /* errmsg: IOCTL is blocked until the device is started */
         RETURN_DEVSTATUS (TAPI_statusIoctlBlocked, 0);
      }

      /* do not use protection for uninitialized device */
      pSemSingleIoctlAccess = IFX_NULL;
   }

   /* retrieve IOCTL handler */
   pHandler = pCtx->pIOCtl->pHandler;
   if (!IFX_TAPI_PtrChk (pHandler.p))
      RETURN_DEVSTATUS (TAPI_statusNotSupported, 0);

   if (TAPI_IS_LL_HANDLER (pHandler))
   {
      /* retrieve low-level handler from driver context */
      pHandler = TAPI_LL_HANDLER_GET (pTapiDev->pDevDrvCtx, pHandler);
      if (!IFX_TAPI_PtrChk (pHandler.p))
         RETURN_DEVSTATUS (TAPI_statusLLNotSupp, 0);

      /* use low-level handle for low-level handler */
      pHnd = pLLHnd;
   }

#if !defined (TAPI_ONE_DEVNODE)
   if (IFX_TAPI_EVENT_GET == pCtx->nCmd)
   {
      /* \note: That hack required to in force set the
         real device number for 'IFX_TAPI_EVENT_GET' service */
      ((IFX_TAPI_EVENT_t*)pArg)->dev = pTapiDev->nDev;
   }
#endif /* !defined (TAPI_ONE_DEVNODE) */

   if (IFX_TAPI_PtrChk (pSemSingleIoctlAccess))
   {
      /* protect ioctl handler against concurrent access */
      TAPI_OS_MutexGet (pSemSingleIoctlAccess);
   }

   /* call IOCtl handler */
   if (TAPI_IOC_CTX_NONE == pCtx->nContext)
   {
      switch (pCtx->pIOCtl->nArgKind)
      {
         case TAPI_ARG_POINTER:
            ret = pHandler.pPtr (pArg);
            break;
         case TAPI_ARG_INTEGER:
            ret = pHandler.pInt ((IFX_uint32_t)pArg);
            break;
         case TAPI_ARG_NONE:
            ret = pHandler.pNone ();
            break;
      }
   }
   else
   {
      switch (pCtx->pIOCtl->nArgKind)
      {
         case TAPI_ARG_POINTER:
            ret = pHandler.pPtrPtr (pHnd, pArg);
            break;
         case TAPI_ARG_INTEGER:
   #ifdef TAPI_ONE_DEVNODE
            if (TAPI_IOC_CTX_CH == pCtx->nContext)
            {
               ret = pHandler.pPtrInt (pHnd, ((TAPI_CH_IO_ARG_t *)pArg)->param);
            }
            else if (TAPI_IOC_CTX_DEV == pCtx->nContext)
            {
               ret = pHandler.pPtrInt (pHnd, ((TAPI_DEV_IO_ARG_t *)pArg)->param);
            }
            else
   #endif /* TAPI_ONE_DEVNODE */
            {
               ret = pHandler.pPtrInt (pHnd, (IFX_uint32_t)pArg);
            }
            break;
         case TAPI_ARG_NONE:
            ret = pHandler.pPtr (pHnd);
            break;
      }
   }

   if (IFX_TAPI_PtrChk (pSemSingleIoctlAccess))
   {
      /* release ioctl handler protection against concurrent access */
      TAPI_OS_MutexRelease (pSemSingleIoctlAccess);
   }

   if (TAPI_IS_LL_HANDLER (pCtx->pIOCtl->pHandler) && !TAPI_SUCCESS (ret))
   {
      /* \note: IFX_TAPI_CAP_CHECK - returns 1 if capability is supported
         IFX_TAPI_RING - returns 1 if phone hooked off
       */
      if ((ret != 1) &&
          ((IFX_TAPI_CAP_CHECK == pCtx->nCmd) ||
           (IFX_TAPI_RING == pCtx->nCmd)))
      {
         /*  correct error code for low-level failure */
         retLL = ret;
         ret = TAPI_statusLLFailed;
      }
   }

   if (!TAPI_SUCCESS (retLL))
      RETURN_DEVSTATUS (ret, retLL);

   return ret;
}
