/******************************************************************************

                              Copyright (c) 2009
                            Lantiq Deutschland GmbH
                     Am Campeon 3; 85579 Neubiberg, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/**
   \file drv_tapi_linux.c
   This file contains the implementation of High-Level TAPI Driver,
   Linux specific part.

   The implementation includes the following parts:
    -Registration part by which the low-level drivers register themselves.
    -Device node operations (open, close, ioctl, read, write, select)
     are done here.
    -Linux module support.
    -Linux /proc fileystem handlers.
    -Timer abstraction layer.
    -Deferring of a function call for later execution.
    -Export of High-Level TAPI function symbols.
*/

/* ============================= */
/* Includes                      */
/* ============================= */
#include "drv_api.h"

#ifdef __KERNEL__
#include <linux/kernel.h>
#endif /* __KERNEL__ */
#ifdef MODULE
#include <linux/module.h>
#endif /* MODULE */

#ifdef __KERNEL__
#include <linux/proc_fs.h>             /*proc-file system*/
#include <linux/timer.h>               /* init_timer() */
#include <linux/init.h>
#include <linux/errno.h>
#include <asm/byteorder.h>
#include <linux/smp_lock.h>         /* lock_kernel() */
#include <asm/io.h>

#ifdef LINUX_2_6
   #include <linux/device.h>
   #include <linux/sched.h>
   #undef   CONFIG_DEVFS_FS
   #ifndef UTS_RELEASE
      /* For newer linux kernels: #include <generated/utsrelease.h> */
      #include <linux/utsrelease.h>
   #endif /* UTC_RELEASE */
#else
   #include <linux/tqueue.h>
   #include <linux/sched.h>
#endif /* LINUX_2_6 */
#endif /* __KERNEL__ */

#include "drv_tapi.h"
#include "drv_tapi_api.h"
#include "drv_tapi_errno.h"
#include "drv_tapi_stream.h"
#include "drv_tapi_ioctl.h"

#ifndef IFXOS_USE_DEV_IO
   #include "drv_tapi_cid.h"
#endif /* IFXOS_USE_DEV_IO */

#include "drv_tapi_ppd.h"
#include "drv_tapi_qos.h"

/* ============================= */
/* Local Macros & Definitions    */
/* ============================= */

/* ================================== */
/* channel specific wrapper structure */
/* ================================== */
#ifndef IFXOS_USE_DEV_IO
typedef struct _TAPI_FD_PRIV_DATA TAPI_FD_PRIV_DATA_t;

struct _TAPI_FD_PRIV_DATA
{
   /* ptr to tapi channel or device specific data */
   IFX_void_t                   *pTapiCtx;
   /* channel fifo number                         */
   IFX_TAPI_STREAM_t            fifo_idx;
};

/* ============================= */
/* kernel ioctl api: structure   */
/* to reference a channel        */
/* (see drv_tapi_kio.h)          */
/* ============================= */
typedef struct
{
   /* storage for major and minor numbers */
   struct inode inode;
   /* storage for private date of real open procedure */
   struct file file;
} IFX_TAPI_Ch_KernRef_t ;

#define MAX_EXPECTED_LL_DRV 5

/* ============================= */
/* Local Functions               */
/* ============================= */
#ifdef __KERNEL__
/*lint -save
   -esym(528,TAPI_timer_call_back)
   -esym(528,TAPI_tqueue) */
static IFX_void_t TAPI_timer_call_back (IFX_ulong_t arg);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static IFX_void_t TAPI_tqueue (IFX_void_t *pWork);
#else /* for Kernel newer or equal 2.6.20 */
static IFX_void_t TAPI_tqueue (struct work_struct *pWork);
#endif /* (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)) */
/*lint -restore */
#endif /* __KERNEL__ */
#ifdef LINUX_2_6
/* TAPI's own RT workqueue */
static struct workqueue_struct *pTapiWq;
/* init struct for the RT workqueue to set the scheduling policy
   via TAPI_DeferWork */
static IFX_TAPI_EXT_EVENT_PARAM_t tapi_wq_setscheduler_param;
#endif /* LINUX_2_6 */

static IFX_int32_t TAPI_SelectCh (TAPI_FD_PRIV_DATA_t *pTapiPriv,
                                  TAPI_OS_drvSelectTable_t *,
                                  TAPI_OS_drvSelectOSArg_t *);

static int ifx_tapi_open (struct inode *inode, struct file *filp);
static int ifx_tapi_release (struct inode *inode, struct file *filp);
static ssize_t ifx_tapi_write(struct file *filp, const char *buf,
                              size_t count, loff_t * ppos);
static ssize_t ifx_tapi_read(struct file * filp, char *buf,
                              size_t length, loff_t * ppos);
static int ifx_tapi_ioctl(struct inode *inode, struct file *filp,
                              unsigned int nCmd, unsigned long nArgument);
static IFX_uint32_t ifx_tapi_poll (struct file *filp, poll_table *table);

#ifdef CONFIG_PROC_FS
static int proc_read_tapi(char *page, char **start, off_t off,
                          int count, int *eof, void *data);
static IFX_int32_t proc_get_tapi_version(IFX_char_t *buf);
static IFX_int32_t proc_get_tapi_status(IFX_char_t *buf);
static IFX_int32_t proc_get_tapi_registered_drivers(IFX_char_t *buf);

extern IFX_int32_t IFX_TAPI_EventWrpBufferPool_ElementCountGet(IFX_void_t);
extern IFX_int32_t IFX_TAPI_EventWrpBufferPool_ElementAvailCountGet(IFX_void_t);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS)
static int proc_read_tapi_stats(char *buf, char **start, off_t offset,
                                int count, int *eof, void *data);
static int proc_write_tapi_stats(struct file *file, const char *buffer,
                                 unsigned long count, void *data);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS) */
static IFX_int32_t proc_read_fifos (IFX_char_t *buf);
static IFX_int32_t proc_EntriesInstall(IFX_void_t);
static IFX_void_t  proc_EntriesRemove(IFX_void_t);
#endif /* CONFIG_PROC_FS */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static IFX_int32_t tapiClassCreate(IFX_void_t);
static IFX_void_t tapiClassRemove(IFX_void_t);
#endif /* >= Linux 2.6.0 */

/*lint -save -esym(752,debug_level) */
/** install parameter debug_level: LOW (1), NORMAL (2), HIGH (3), OFF (4) */
#ifdef ENABLE_TRACE
extern IFX_uint32_t debug_level;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17))
MODULE_PARM(debug_level, "i");
#else
module_param(debug_level, uint, 0);
#endif /* < 2.6.17 */
MODULE_PARM_DESC(debug_level, "set to get more (1) or fewer (4) debug outputs");
#endif /* ENABLE_TRACE */
/*lint -restore */

IFX_int32_t block_egress_tasklet = 0;
IFX_int32_t block_ingress_tasklet = 0;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17))
MODULE_PARM(block_egress_tasklet, "i");
MODULE_PARM(block_ingress_tasklet, "i");
#else
module_param(block_egress_tasklet, int, 0);
module_param(block_ingress_tasklet, int, 0);
#endif /* < 2.6.17 */
MODULE_PARM_DESC(block_egress_tasklet, "block the registration of egress tasklets, i.e. force to use the RT kernel thread");
MODULE_PARM_DESC(block_ingress_tasklet, "block the execution of the ingress tasklet, i.e. force to use the RT kernel thread");

/** The driver callbacks which will be registered with the kernel*/
static struct file_operations tapi_fops;

/* ============================= */
/* Local variables               */
/* ============================= */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static struct class *pTAPI_Class;
#endif /* >= Linux 2.6.0 */

/* ============================= */
/* Global function definition    */
/* ============================= */

/**
   Register the low level driver with the operating system.

   This function will be called when the low-level driver is added to the
   system. Any OS specific registration or set-up should be done here.

   For Linux kernels with devfs support device nodes will be created here.
   The number of nodes created depend on the "one device node" or "multiple
   device node" configuration.
   For Linux kernels without devfs support the character driver is registered.

   \param  pLLDrvCtx    Pointer to device driver context created by LL-driver.
   \param  pHLDrvCtx    Pointer to high-level driver context.

   \return
   TAPI_statusOk

   \remarks
   For Linux the device nodes need to be registered and then the driver
   itself is registered with the kernel.
*/
IFX_return_t TAPI_OS_RegisterLLDrv (IFX_TAPI_DRV_CTX_t* pLLDrvCtx,
                                    IFX_TAPI_HL_DRV_CTX_t* pHLDrvCtx)
{
   IFX_int32_t ret = 0;
   IFX_uint16_t majorNumber;
#if defined (CONFIG_DEVFS_FS) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
   IFX_uint16_t minorNumber;
   IFX_uint16_t
      maxDevices = 1,
      maxChannels = 0,
      nDevIdx = 0,
      nChIdx = 0;
   IFX_char_t sDevName[15];
   IFX_uint16_t nTapiFdIdx = 0;
#endif /* defined (CONFIG_DEVFS_FS) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)) */

   if (tapi_fops.ioctl == IFX_NULL)
   {
#ifdef MODULE
      tapi_fops.owner =    THIS_MODULE;
#endif
      tapi_fops.read =     ifx_tapi_read;
      tapi_fops.write =    ifx_tapi_write;
      tapi_fops.poll =     ifx_tapi_poll;
      tapi_fops.ioctl =    ifx_tapi_ioctl;
      tapi_fops.open =     ifx_tapi_open;
      tapi_fops.release =  ifx_tapi_release;
   }

   /* copy registration info from Low level driver */
   majorNumber = pLLDrvCtx->majorNumber;

   /* limit devNodeName to 8 characters */
   sprintf (pHLDrvCtx->registeredDrvName, "ifx_tapi (%.8s)", pLLDrvCtx->devNodeName);

#if  (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
   /* Register the character device */
   ret = register_chrdev (majorNumber, pHLDrvCtx->registeredDrvName, &tapi_fops);
#else /* < Linux 2.6.0 */
   ret = register_chrdev_region(MKDEV(majorNumber, 0), 1, pHLDrvCtx->registeredDrvName);
#endif /* < Linux 2.6.0 */
   if (ret < 0)
   {
      TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
           ("IFX_TAPI_Register_LL_Drv: unable to register chrdev, major number "
            "%d\n\r", majorNumber));
      return TAPI_statusErr;
   }

#if defined (CONFIG_DEVFS_FS) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#if !defined (TAPI_ONE_DEVNODE)
   maxDevices  = pLLDrvCtx->maxDevs;
   maxChannels = pLLDrvCtx->maxChannels;

   /* limit to one digit , more than 10 is not possible with the multi device
      node concept fd10 .. fd19 */
   if (maxChannels > 9)
   {
      TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
         ("TAPI_DRV (major:%d): The channel resources limited up to 9, "
         "due to device entry naming limitation.\n",
         majorNumber));
      maxChannels = 9;
   }
#endif /* not TAPI_ONE_DEVNODE */

   for (nDevIdx = 0; nDevIdx < maxDevices; nDevIdx++)
   {
      for (nChIdx = 0; nChIdx <= maxChannels; nChIdx++)
      {
#if defined (TAPI_ONE_DEVNODE)
         minorNumber = 0;
         /* limit devNodeName to 8 characters */
         snprintf (sDevName, sizeof(sDevName), "%.8s",
            pLLDrvCtx->devNodeName);
#else /* TAPI_ONE_DEVNODE */
         /* First device have index 1 */
         minorNumber = pLLDrvCtx->minorBase * (nDevIdx + 1);
         minorNumber += nChIdx;
         /* limit devNodeName to 8 characters */
         snprintf (sDevName, sizeof(sDevName), "%.8s%d",
            pLLDrvCtx->devNodeName, minorNumber);
#endif /* TAPI_ONE_DEVNODE */

         if (nTapiFdIdx >= TAPI_MAX_DEVFS_HANDLES)
         {
            TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
                 ("TAPI_DRV: Stopped character device creation on '%s'. "
                  "Reached maximal number(nTapiFdIdx=%d) of file "
                  "descriptors supported by TAPI\n",
                  sDevName, nTapiFdIdx));
            /* terminate node creation */
            nDevIdx = maxDevices;
            break;
         }

#ifdef CONFIG_DEVFS_FS
         /* Register the device node with the kernel */
         pHLDrvCtx->TAPI_devfs_handle[nTapiFdIdx] =
            devfs_register(IFX_NULL, sDevName, DEVFS_FL_DEFAULT,
            majorNumber, minorNumber, (S_IFCHR|S_IRUGO|S_IWUGO),
            &tapi_fops , (void *)0);
         if (IFX_NULL == pHLDrvCtx->TAPI_devfs_handle[nTapiFdIdx])
         {
            TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
                 ("IFX_TAPI_Register_LL_Drv: unable to add device /dev/%s to "
                  "dev_fs\n\r",
                  sDevName));
            return TAPI_statusErr;
         }
#endif /* CONFIG_DEVFS_FS */

#if  (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
         /* add character device */
         pHLDrvCtx->pTAPI_cdev[nTapiFdIdx] = cdev_alloc();
         if (IFX_NULL == pHLDrvCtx->pTAPI_cdev[nTapiFdIdx])
         {
            TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
                 ("IFX_TAPI_Register_LL_Drv: unable to allocate memory "
                 "for character device '%s'\n\r", sDevName));
            return TAPI_statusNoMem;
         }

         cdev_init(pHLDrvCtx->pTAPI_cdev[nTapiFdIdx], &tapi_fops);
         pHLDrvCtx->pTAPI_cdev[nTapiFdIdx]->owner = THIS_MODULE;
         pHLDrvCtx->pTAPI_cdev[nTapiFdIdx]->ops = &tapi_fops;

         ret = cdev_add(pHLDrvCtx->pTAPI_cdev[nTapiFdIdx],
            MKDEV(majorNumber, minorNumber), 1);
         if (ret != 0)
         {
            TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
                 ("IFX_TAPI_Register_LL_Drv: unable to add chrdev "
                  "%s\n\r", sDevName));

            cdev_del (pHLDrvCtx->pTAPI_cdev[nTapiFdIdx]);
            pHLDrvCtx->pTAPI_cdev[nTapiFdIdx] = IFX_NULL;
            return TAPI_statusErr;
         }

         pHLDrvCtx->pLL_Device[nTapiFdIdx] =
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
            class_device_create(pTAPI_Class, MKDEV(majorNumber, minorNumber),
                                IFX_NULL, "%s", sDevName);
#else
            device_create(pTAPI_Class, IFX_NULL,
                          MKDEV(majorNumber, minorNumber),
                          "%s", sDevName);
#endif /* < Linux 2.6.18 */
         if (IS_ERR(pHLDrvCtx->pLL_Device[nTapiFdIdx]))
         {
            TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
                 ("IFX_TAPI_Register_LL_Drv: unable to create class device "
                  "'%s'\n\r", sDevName));
            return TAPI_statusErr;
         }
#endif /* >= Linux 2.6.0 */
         nTapiFdIdx++;
      }
   }
#endif /* defined (CONFIG_DEVFS_FS) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)) */

   return TAPI_statusOk;
}

/**
   UnRegister the low level driver from the system.

   This function will be called when the low-level driver is removed from the
   system. Any OS specific deregistration should be done here.

   For Linux kernels with devfs support device nodes will be unregistered here.
   The number of nodes unregistered depend on the "one device node" or "multiple
   device node" configuration.
   For Linux kernels without devfs support the character driver is unregistered.

   \param  pLLDrvCtx    Pointer to device driver context created by LL-driver.
   \param  pHLDrvCtx    Pointer to high-level driver context.

   \return
   TAPI_statusErr on error, otherwise TAPI_statusOk
*/
IFX_return_t TAPI_OS_UnregisterLLDrv (IFX_TAPI_DRV_CTX_t* pLLDrvCtx,
                                      IFX_TAPI_HL_DRV_CTX_t* pHLDrvCtx)
{
#if defined (CONFIG_DEVFS_FS) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
   int i;

   for (i=0; i < TAPI_MAX_DEVFS_HANDLES; ++i)
   {
#ifdef CONFIG_DEVFS_FS
      /* remove vinetic devices from dev fs */
      if (pHLDrvCtx->TAPI_devfs_handle[i])
      {
         devfs_unregister(pHLDrvCtx->TAPI_devfs_handle[i]);
         pHLDrvCtx->TAPI_devfs_handle[i] = IFX_NULL;
      }
#endif /* CONFIG_DEVFS_FS */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
      if (pHLDrvCtx->pLL_Device[i])
      {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
         class_device_destroy(pTAPI_Class, pHLDrvCtx->pLL_Device[i]->devt);
#else
         device_destroy(pTAPI_Class, pHLDrvCtx->pLL_Device[i]->devt);
#endif /* < Linux 2.6.18 */
         pHLDrvCtx->pLL_Device[i] = IFX_NULL;
      }

      if (pHLDrvCtx->pTAPI_cdev[i])
      {
         cdev_del(pHLDrvCtx->pTAPI_cdev[i]);
         pHLDrvCtx->pTAPI_cdev[i] = IFX_NULL;
      }
#endif /* >= Linux 2.6.0 */
   }
#endif /* defined (CONFIG_DEVFS_FS) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)) */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
   unregister_chrdev (pLLDrvCtx->majorNumber, pHLDrvCtx->registeredDrvName);
#else
   unregister_chrdev_region (MKDEV(pLLDrvCtx->majorNumber, 0), 1);
#endif

   return TAPI_statusOk;
}

/**
   Open the device.

   At the first time:
   - Initialize the high-level TAPI device structure
   - Call the low-level function to initialise the low-level device structure
   - Initialize the high-level TAPI channel structure
   - Call the low-level function to initialise the low-level channel structure

   \param  inode        Pointer to the inode.
   \param  filp         Pointer to the file descriptor.

   \return
   0 - if no error,
   otherwise error code
*/
static int ifx_tapi_open (struct inode *inode, struct file *filp)
{
   TAPI_FD_PRIV_DATA_t     *pTapiPriv  = IFX_NULL;
   IFX_TAPI_DRV_CTX_t      *pDrvCtx    = IFX_NULL;
   TAPI_DEV                *pTapiDev   = IFX_NULL;
#if !defined(TAPI_ONE_DEVNODE)
   TAPI_CHANNEL            *pTapiCh    = IFX_NULL;
#endif /* not TAPI_ONE_DEVNODE */
   IFX_uint32_t            nDev = 0,
                           nCh = 0;
   IFX_uint32_t            majorNum = 0;
   IFX_uint32_t            minorNum = 0;

   majorNum = MAJOR(inode->i_rdev);
   minorNum = MINOR(inode->i_rdev);
   TRACE (TAPI_DRV, DBG_LEVEL_LOW,
         ("ifxTAPI open %d/%d\n", majorNum, minorNum));

   /* Get the pointer to the device driver context based on the major number */
   pDrvCtx = IFX_TAPI_DeviceDriverContextGet (majorNum);

   if (pDrvCtx == IFX_NULL)
   {
      /* This should actually never happen because the file descriptors are
         registered in TAPI_OS_RegisterLLDrv after a driver context is known. */
      printk (KERN_INFO "tapi_open: error finding DrvCtx\n");
      return -ENODEV;
   }

#if !defined(TAPI_ONE_DEVNODE)
   /* calculate the device number and channel number */
   nDev = ((IFX_uint32_t)minorNum / pDrvCtx->minorBase) - 1;
   nCh   = ((IFX_uint32_t)minorNum % pDrvCtx->minorBase);

   /* check the device number */
   if (nDev >= pDrvCtx->maxDevs)
   {
      TRACE(TAPI_DRV, DBG_LEVEL_HIGH,
           ("TAPI_DRV: max. device number exceed\n"));
      return -ENODEV;
   }

   /* check the channel number */
   if (nCh > pDrvCtx->maxChannels)
   {
      TRACE(TAPI_DRV, DBG_LEVEL_HIGH,
          ("TAPI_DRV: max. channel number exceed\n"));
      return -ENODEV;
   }
#endif /* not TAPI_ONE_DEVNODE */

   pTapiDev = pDrvCtx->pTapiDev + nDev;

   /* Allocate memory for a new TAPI context wrapper structure */
   pTapiPriv = (TAPI_FD_PRIV_DATA_t* )
               TAPI_OS_Malloc (sizeof (TAPI_FD_PRIV_DATA_t));

   if (pTapiPriv == IFX_NULL)
   {
      TRACE(TAPI_DRV, DBG_LEVEL_HIGH,
            ("Unable to allocate memory for TAPI private data. "
             "[dev:%d ch:%d]\n", nDev, nCh));
      return -ENOMEM;
   }

#if !defined(TAPI_ONE_DEVNODE)
   if (nCh == 0)
#endif /* not TAPI_ONE_DEVNODE */
   {
      pTapiPriv->pTapiCtx = pTapiDev;
   }
#if !defined(TAPI_ONE_DEVNODE)
   else
   {
      pTapiCh = pTapiDev->pTapiChanelArray + nCh - 1;

      pTapiPriv->pTapiCtx = pTapiCh;
   }
#endif /* not TAPI_ONE_DEVNODE */

   pTapiPriv->fifo_idx = IFX_TAPI_STREAM_COD;
   filp->private_data = pTapiPriv;

   /* Call the Low level Device specific open routine */
   if (IFX_TAPI_PtrChk (pDrvCtx->Open))
   {
      IFX_int32_t retLL;

      retLL = pDrvCtx->Open (pTapiPriv->pTapiCtx);

      if (!TAPI_SUCCESS(retLL))
      {
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH,
               ("Open LL channel failed. [dev:%d ch:%d]\n", nDev, nCh));

         TAPI_OS_Free (pTapiPriv);

         filp->private_data = IFX_NULL;
         return -ENODEV;
      }
   }

   /* increment the use counter */
   pTapiDev->nInUse++;

#if !defined(TAPI_ONE_DEVNODE)
   /* increment the use counters */
   if (IFX_NULL != pTapiCh)
   {
      pTapiCh->nInUse++;
   }
#endif /* not TAPI_ONE_DEVNODE */

   /* increment module use counter */
#ifdef MODULE
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
   MOD_INC_USE_COUNT;
#endif /* (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)) */
#endif /* MODULE */

   return 0;
}


/**
   Close a file.

   This function gets called when a close is called on the file descriptor.
   Both types device and channel file descriptors are handled here. The
   function decrements the usage count, and calls the LL release routine.

   \param  inode        Pointer to the inode.
   \param  filp         Pointer to the file descriptor.

   \return
   0 - if no error,
   otherwise error code
*/
static int ifx_tapi_release (struct inode *inode, struct file *filp)
{
   /* nCh is the first field in both TAPI_DEV and TAPI_CHANNEL */
   TAPI_FD_PRIV_DATA_t *pTapiPriv = IFX_NULL;
   IFX_uint8_t          nCh = 0;
   IFX_TAPI_DRV_CTX_t  *pDrvCtx = IFX_NULL;
   TAPI_DEV            *pTapiDev = IFX_NULL;
   TAPI_CHANNEL        *pTapiCh = IFX_NULL;
   IFX_TAPI_LL_CH_t    *pLLChDev = IFX_NULL;

   if (IFX_NULL == filp->private_data)
   {
      /* device was already released */
      return -ENODEV;
   }

   pTapiPriv = (TAPI_FD_PRIV_DATA_t* )filp->private_data;
   nCh = ((TAPI_DEV *) pTapiPriv->pTapiCtx)->nChannel;

   TRACE(TAPI_DRV, DBG_LEVEL_LOW, ("ifxTAPI close %d/%d tapi-ch %d\n\r",
         MAJOR(inode->i_rdev), MINOR(inode->i_rdev), nCh));

   if (nCh != IFX_TAPI_DEVICE_CH_NUMBER)
   {
      /* closing channel file descriptor */
      pTapiCh = (TAPI_CHANNEL *) pTapiPriv->pTapiCtx;

      pTapiDev = pTapiCh->pTapiDevice;
      pLLChDev = pTapiCh->pLLChannel;

   }
   else
   {
      /* closing device file descriptor */
      pTapiDev = (TAPI_DEV *) pTapiPriv->pTapiCtx;
      pLLChDev = pTapiDev->pLLDev;
   }

   pDrvCtx = pTapiDev->pDevDrvCtx;

   /* Call the Low-level Device specific release routine. */
   /* Not having such a function is not an error. */
   if (IFX_TAPI_PtrChk (pDrvCtx->Release))
   {
      IFX_int32_t retLL;

      retLL = pDrvCtx->Release (pLLChDev);

      if (!TAPI_SUCCESS (retLL))
      {
         TRACE(TAPI_DRV, DBG_LEVEL_HIGH,
               ("Release LL channel failed for ch: %d\n", nCh));
         return -EBUSY;
      }
   }

   /* decrement the use counters */
   if ((IFX_NULL != pTapiCh) && (pTapiCh->nInUse > 0))
   {
      pTapiCh->nInUse--;
   }

   /* decrement the use counter */
   if (pTapiDev->nInUse > 0)
   {
      pTapiDev->nInUse--;
   }

   /* decrement use counter */
#ifdef MODULE
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
   MOD_DEC_USE_COUNT;
#endif /* (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)) */
#endif /* MODULE */

   /* free private data */
   TAPI_OS_Free (pTapiPriv);

   filp->private_data = IFX_NULL;

   return 0;
}


/**
   Writes data to the device.

   \param  filp         Pointer to the file descriptor.
   \param  buf          Pointer to data to be written.
   \param  count        Data length.
   \param  ppos         unused.

   \return
   length or a negative error code
   \remarks
   Currently the low-level write function is called transparently
*/
static ssize_t ifx_tapi_write (struct file *filp, const char *buf,
                               size_t count, loff_t * ppos)
{
#if (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET)
   TAPI_FD_PRIV_DATA_t *pTapiPriv = (TAPI_FD_PRIV_DATA_t* )filp->private_data;
   TAPI_CHANNEL        *pTapiCh = (TAPI_CHANNEL* )pTapiPriv->pTapiCtx;
   TAPI_DEV            *pTapiDev  = pTapiCh->pTapiDevice;
   IFX_TAPI_DRV_CTX_t  *pDrvCtx   = pTapiDev->pDevDrvCtx;
   IFX_TAPI_STREAM_t    fifo_idx  = pTapiPriv->fifo_idx;
   IFX_uint8_t         *pData;
   IFX_uint32_t         buf_size;
   IFX_ssize_t          size = 0;

   if (pTapiDev->bInitialized == IFX_FALSE)
   {
      TRACE(TAPI_DRV, DBG_LEVEL_LOW,
            ("TAPI_DRV: write failed because device is not initialised\n"));
      return -EFAULT;
   }


   if (!IFX_TAPI_PtrChk (pDrvCtx->Write))
   {
      IFX_TAPI_Stat_Add(pTapiCh, fifo_idx,
                        TAPI_STAT_COUNTER_INGRESS_DISCARDED, 1);
      TRACE(TAPI_DRV, DBG_LEVEL_LOW,
            ("TAPI_DRV: LL-driver does not provide packet write\n"));
      return -EFAULT;
   }

   /* Truncate data to the size of our voice buffer. */
   buf_size = IFX_TAPI_VoiceBufferPool_ElementSizeGet();
   buf_size -= pDrvCtx->pktBufPrependSpace;
   if ((IFX_uint32_t)count < buf_size)
   {
      buf_size = (IFX_uint32_t)count;
   }

   /* Get a buffer from the voice buffer pool. */
   if ((pData = IFX_TAPI_VoiceBufferGet()) == IFX_NULL)
   {
      /* voice buffer pool is depleted */
      IFX_TAPI_Stat_Add(pTapiCh, fifo_idx,
                        TAPI_STAT_COUNTER_INGRESS_DISCARDED, 1);
      return -EFAULT;
   }

   /* if you have time and want clean buffers you might like this */
   /* memset(pData, 0, buf_size); */

   /* Copy data from userspace into kernelspace buffer. */
   if (!TAPI_OS_CpyUsr2Kern (
         (void *)(pData + pDrvCtx->pktBufPrependSpace),
         (void *)buf, buf_size))
   {
      /* copy into kernel space failed */
      IFX_TAPI_Stat_Add(pTapiCh, fifo_idx,
                        TAPI_STAT_COUNTER_INGRESS_DISCARDED, 1);
      IFX_TAPI_VoiceBufferPut(pData);
      return -EFAULT;
   }

   /* Call the Low level driver's write function. */
   size = pDrvCtx->Write (pTapiCh->pLLChannel, pData, (IFX_int32_t)buf_size,
                          (IFX_int32_t*)(IFX_void_t*)ppos, fifo_idx);

   if (size == IFX_ERROR)
   {
      /* Low level driver's write failed */
      IFX_TAPI_Stat_Add(pTapiCh, fifo_idx,
                        TAPI_STAT_COUNTER_INGRESS_DISCARDED, 1);
      IFX_TAPI_VoiceBufferPut(pData);
      return -EFAULT;
   }

   return (ssize_t) size;
#else
   IFX_UNUSED (filp);
   IFX_UNUSED (buf);
   IFX_UNUSED (count);
   IFX_UNUSED (ppos);

   return (ssize_t) 0;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET) */
}


/**
   Reads data from the device.

   \param  filp         Pointer to the file descriptor.
   \param  buf          Pointer to buffer that receives the data.
   \param  count        Max size of data to read in bytes.
   \param  ppos         unused.

   \return
   Number of bytes read successfully into the buffer.
*/
static ssize_t ifx_tapi_read(struct file *filp,
                             char *buf, size_t count, loff_t * ppos)
{
#if (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET)
   TAPI_FD_PRIV_DATA_t *pTapiPriv = (TAPI_FD_PRIV_DATA_t* )filp->private_data;
   TAPI_CHANNEL        *pTapiCh = (TAPI_CHANNEL* )pTapiPriv->pTapiCtx;
   TAPI_DEV            *pTapiDev  = pTapiCh->pTapiDevice;
   IFX_TAPI_DRV_CTX_t  *pDrvCtx   = pTapiDev->pDevDrvCtx;
   IFX_TAPI_STREAM_t    fifo_idx  = pTapiPriv->fifo_idx;
   IFX_void_t const    *pPacket   = NULL;
   IFX_uint32_t         nOffset;
   IFX_uint32_t         size = 0;

   IFX_UNUSED (ppos);

   if (pTapiDev->bInitialized == IFX_FALSE)
   {
      TRACE(TAPI_DRV, DBG_LEVEL_LOW,
            ("TAPI_DRV: read failed because device is not initialised\n"));
      return -EFAULT;
   }

   if (fifo_idx >= IFX_TAPI_STREAM_MAX)
   {
      /* index out of range - internal error */
      TRACE (TAPI_DRV, DBG_LEVEL_LOW, ("INFO: FIFO index out of range\n"));
      /* return linux error code to the application */
      return -EINVAL;
   }

   /* All packets to be read are already in the channel specific upstream
      fifo of the driver. They are put there by whoever receives packets.
      So all that is done here operates on this fifo and does not need to call
      any ll-driver function any more. */

   /* check for non blocking flag */
   if (filp->f_flags & O_NONBLOCK)
      pTapiCh->nFlags |= CF_NONBLOCK;

   /* Lock access to the fifo. */
   TAPI_OS_MutexGet (&pTapiCh->semTapiChDataLock);
   if (IFX_TAPI_PtrChk (pDrvCtx->IRQ.LockDevice))
   {
      pDrvCtx->IRQ.LockDevice (pTapiDev->pLLDev);
   }

   /* If FIFO is empty or does not exist we cannot read anything. In this
      case we either wait until data arrives or return if non-blocking.
      If the FIFO does not exist no data will ever arrive and the possible
      wait will last forever. */
   if ((pTapiCh->pUpStreamFifo[fifo_idx] == IFX_NULL) ||
       (fifoEmpty(pTapiCh->pUpStreamFifo[fifo_idx]) == IFX_TRUE))
   {
      /* Unlock access to the fifo. */
      if (pDrvCtx->IRQ.UnlockDevice != IFX_NULL)
      {
         pDrvCtx->IRQ.UnlockDevice (pTapiDev->pLLDev);
      }
      TAPI_OS_MutexRelease (&pTapiCh->semTapiChDataLock);

      /* check the non-blocking flag */
      if (pTapiCh->nFlags & CF_NONBLOCK)
      {
         /* this is a non blocking read call - return immediately */
         return 0;
      }

      /* this is a blocking read call so go to sleep now */
      if (TAPI_OS_EventWait (&pTapiCh->semReadBlock,
                             TAPI_OS_WAIT_FOREVER, IFX_NULL) == IFX_ERROR)
      {
         /* timeout has expired without arrival of data */
         return 0;
      }

      /* we have been woken because data has arrived */

      /* before accessing the mailbox - lock again. */
      TAPI_OS_MutexGet (&pTapiCh->semTapiChDataLock);
      if (IFX_TAPI_PtrChk (pDrvCtx->IRQ.LockDevice))
      {
         pDrvCtx->IRQ.LockDevice (pTapiDev->pLLDev);
      }
   }

   /* read data from the fifo */
   TAPI_ASSERT (fifoEmpty(pTapiCh->pUpStreamFifo[fifo_idx]) ==
                 IFX_FALSE);
   pPacket = IFX_TAPI_UpStreamFifo_Get(pTapiCh, fifo_idx, &size, &nOffset);

   /* Unlock access to the fifo. */
   if (pDrvCtx->IRQ.UnlockDevice != IFX_NULL)
   {
      pDrvCtx->IRQ.UnlockDevice (pTapiDev->pLLDev);
   }
   TAPI_OS_MutexRelease (&pTapiCh->semTapiChDataLock);

   /* sanity check on the received packet */
   if (pPacket == NULL)
   {
      /* packet may not be NULL - internal error */
      TRACE (TAPI_DRV, DBG_LEVEL_LOW, ("INFO: pPacket Null\n"));
      return 0;
   }
   if (size > (IFX_uint32_t)count)
   {
      /* output buffer not large enough for data */
      /* drop the packet */
      IFX_TAPI_VoiceBufferPut((IFX_void_t *)pPacket);
      /* update statistic */
      IFX_TAPI_Stat_Add(pTapiCh, fifo_idx,
                        TAPI_STAT_COUNTER_EGRESS_DISCARDED, 1);
      /* return linux error code to the application */
      return -EINVAL;
   }

   /* unmap data */
   TAPI_OS_CpyKern2Usr (buf, ((IFX_char_t *)pPacket + nOffset),
                        (IFX_uint32_t)size);

   /* return buffer back to the pool now that is has been copied */
   IFX_TAPI_VoiceBufferPut((IFX_void_t *)pPacket);
   /* update statistic */
   IFX_TAPI_Stat_Add(pTapiCh, fifo_idx,
                     TAPI_STAT_COUNTER_EGRESS_DELIVERED, 1);

   return (ssize_t) size;
#else
   IFX_UNUSED (filp);
   IFX_UNUSED (buf);
   IFX_UNUSED (count);
   IFX_UNUSED (ppos);

   return (ssize_t) 0;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET) */
}

/**
   Poll implementation for the device.

   \param file pointer to the file descriptor
   \param wait pointer to the poll table

   \return
    status of the events occured on the device which are
    0, TAPI_OS_SYSEXCEPT, TAPI_OS_SYSREAD, TAPI_OS_SYSWRITE
   \remarks
   This function does the following functions:
      - Put the event queue in the wait table and sleep if select
        is called on the device itself.
      - Put the read/write queue in the wait table and sleep if select
        is called on the channels
*/
static IFX_uint32_t ifx_tapi_poll (struct file *filp, poll_table *wait)
{
   TAPI_FD_PRIV_DATA_t *pTapiPriv = (TAPI_FD_PRIV_DATA_t* )filp->private_data;
   TAPI_DEV            *pTapiDev  = (TAPI_DEV* ) pTapiPriv->pTapiCtx;
   IFX_uint16_t         i;
#ifdef TAPI_ONE_DEVNODE
   IFX_uint16_t         j;
   IFX_TAPI_DRV_CTX_t*  pDrvCtx = IFX_NULL;
#endif /* TAPI_ONE_DEVNODE */
   IFX_int32_t          ret = 0;

   /* check device ptr */
   if (pTapiDev == IFX_NULL)
   {
      return (IFX_uint32_t)(-ENODEV);
   }

#ifdef TAPI_ONE_DEVNODE
   /* get driver context in case select called on device file descriptor */
   if (pTapiDev->nChannel == IFX_TAPI_DEVICE_CH_NUMBER)
   {
      pDrvCtx = IFX_TAPI_DeviceDriverContextGet (
         pTapiDev->pDevDrvCtx->majorNumber);
   }

   if (pTapiDev->nChannel == IFX_TAPI_DEVICE_CH_NUMBER &&
       pDrvCtx != IFX_NULL)
   {
      for (j= 0;j < pDrvCtx->maxDevs; ++j)
      {
         pTapiDev = &pDrvCtx->pTapiDev[j];

         TAPI_OS_DrvSelectQueueAddTask (filp, &pTapiDev->wqEvent, wait);

         if (pTapiDev->pTapiChanelArray == IFX_NULL)
            continue;

         for (i = 0; i < pTapiDev->nMaxChannel; i++)
         {
            TAPI_CHANNEL *pTapiCh = &(pTapiDev->pTapiChanelArray[i]);

            TAPI_OS_DrvSelectQueueAddTask (filp, &pTapiCh->semReadBlock.object,
                                           wait);
            TAPI_OS_DrvSelectQueueAddTask (filp, &pTapiCh->wqRead, wait);

            if ((pTapiCh->bInitialized == IFX_TRUE) &&
                (IFX_TAPI_EventFifoEmpty(pTapiCh) == IFX_FALSE))
            {
               /* exception available so return action */
               ret |= TAPI_OS_SYSREAD;
               break;
            }
         }
      }
   }
   else
#endif /* TAPI_ONE_DEVNODE */
   {
      if (pTapiDev->nChannel == IFX_TAPI_DEVICE_CH_NUMBER)
      {
         /* Register the TAPI-event waitqueue as wakeup source. */
         TAPI_OS_DrvSelectQueueAddTask (filp, &pTapiDev->wqEvent, wait);

         /* Check if there is any TAPI-event on any of the TAPI channels and
            return if file operations are possible without blocking. */

         TAPI_ASSERT (pTapiDev->pTapiChanelArray != IFX_NULL);

         for (i = 0; i < pTapiDev->nMaxChannel; i++)
         {
            TAPI_CHANNEL *pTapiCh = &(pTapiDev->pTapiChanelArray[i]);
            if ((pTapiCh->bInitialized == IFX_TRUE) &&
                (IFX_TAPI_EventFifoEmpty(pTapiCh) == IFX_FALSE))
            {
               /* TAPI-event available so return action */
               ret |= TAPI_OS_SYSREAD;
               break;
            }
         }
      }
      else if (pTapiDev->nChannel < pTapiDev->nMaxChannel)
      {
         /* Check if any packet is available in any of the upstream fifos and
            return if file operations are possible without blocking. */
         ret |= TAPI_SelectCh (pTapiPriv,
                               (TAPI_OS_drvSelectTable_t *)wait,
                               (TAPI_OS_drvSelectOSArg_t *)filp);
      }
   }

   return (IFX_uint32_t)ret;
}

/**
   Configuration / Control for the device.

   \param  inode        Pointer to the inode.
   \param  filp         Pointer to the file descriptor.
   \param  nCmd         IOCTL identifier.
   \param  nArg         Optional argument.

   \return
   0 and positive values - success,
   negative value - ioctl failed
   \remarks
   This function does the following functions:
      - If the ioctl command is device specific, low-level driver's ioctl function
      - If the ioctl command is TAPI specific, it is handled at this level
*/
static int ifx_tapi_ioctl(struct inode *inode, struct file *filp,
                          unsigned int nCmd, unsigned long nArg)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx = IFX_NULL;
   IFX_TAPI_ioctlCtx_t ctx;
   IFX_int32_t ret;

   IFX_UNUSED (filp);

   /* get the device driver context */
   pDrvCtx = IFX_TAPI_DeviceDriverContextGet (MAJOR(inode->i_rdev));
   if (pDrvCtx == IFX_NULL)
   {
      return IFX_ERROR;
   }

   /* get the ioctl context: channel, device etc. */
   ret = TAPI_ioctlContextGet (pDrvCtx, MINOR(inode->i_rdev),
      nCmd, nArg, IFX_TRUE, &ctx);
   if (IFX_SUCCESS == ret)
   {
      ret = TAPI_Ioctl (&ctx);
   }

   TAPI_ioctlContextPut (nArg, &ctx);

   return ret;
}

#ifdef CONFIG_PROC_FS
/**
   Read the version information from the driver.

   \param  buf          Destination buffer.

   \return
   Number of bytes written into the buffer.
*/
static IFX_int32_t proc_get_tapi_version(IFX_char_t *buf)
{
    IFX_int32_t len;

    len = sprintf(buf, "%s\n", &TAPI_WHATVERSION[4]);
    len += sprintf(buf + len, "Compiled on %s, %s for Linux kernel %s\n",
                   __DATE__, __TIME__, UTS_RELEASE);

    return len;
}


#ifdef HAVE_CONFIG_H
/**
   Read the configure parameters of the driver.

   \param  buf    Destination buffer.

   \return
   Length of the string returned in buf.
*/
static IFX_int32_t proc_ConfigureGet (IFX_char_t *buf)
{
   IFX_int32_t len;

   len  = sprintf(buf, "configure %s\n", &DRV_TAPI_WHICHCONFIG[0]);

   return len;
}
#endif /* HAVE_CONFIG_H */


/**
   Read the status information from the driver.

   \param  buf          Destination buffer.

   \return
   Number of bytes written into the buffer.
*/
static IFX_int32_t proc_get_tapi_status(IFX_char_t *buf)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx  = IFX_NULL;
   TAPI_DEV           *pTapiDev = IFX_NULL;
   IFX_int32_t len=0;
   IFX_int32_t i, j;

   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      if (gHLDrvCtx [i].pDrvCtx == IFX_NULL)
      {
         continue;
      }

      pDrvCtx = gHLDrvCtx [i].pDrvCtx;
      len += sprintf(buf+len,
         "-- %s --\n", pDrvCtx->drvName);

      if (pDrvCtx->pTapiDev == IFX_NULL)
      {
         /* driver not initialised yet -> skip it */
         break;
      }

      for (j = 0; j < pDrvCtx->maxDevs; j++)
      {
         pTapiDev = &(pDrvCtx->pTapiDev[j]);
         len += sprintf(buf+len,
                "  Device #%d: Initialized = %d\n",
                 j, pTapiDev->bInitialized);
      }
   }
   return len;
}


/**
   Read the registered low level driver information

   \param  buf          Destination buffer.

   \return
   Number of bytes written into the buffer.
*/
static IFX_int32_t proc_get_tapi_registered_drivers(IFX_char_t *buf)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx = IFX_NULL;
   IFX_int32_t i, len = 0;

   len += sprintf(buf+len, "%-10s %-10s %7s %10s %-10s\n",
       "LL-Driver", "version", "major", "devices", "devname");
   len += sprintf(buf+len,
      "==================================================\n");

   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      if (gHLDrvCtx [i].pDrvCtx != IFX_NULL)
      {
         pDrvCtx = gHLDrvCtx [i].pDrvCtx;
         len += sprintf(buf+len, "%-10s %-10s %7d %10d /dev/%-10s\n",
              pDrvCtx->drvName, pDrvCtx->drvVersion,
              pDrvCtx->majorNumber,pDrvCtx->maxDevs, pDrvCtx->devNodeName);
      }
   }

#if (TAPI_CFG_FEATURES & TAPI_FEAT_QOS)
   /* Print details about the QOS driver that is registered with TAPI. */
   len += sprintf(buf+len, "\nQOS driver:\n");
   len += IFX_TAPI_QOS_proc_read_registration(buf+len);
#endif /* defined(LINUX) && defined(CONFIG_PROC_FS) */

   return len;
}

/**
   Read the current bufferpool status

   \param  buf          Destination buffer.

   \return
   Number of bytes written into the buffer.
*/
static IFX_int32_t proc_read_bufferpool (IFX_char_t *buf)
{
   IFX_int32_t len = 0;

   len += sprintf(buf+len,
                  "TAPIpool free/total voice (%4d/%4d), "
#if (TAPI_CFG_FEATURES & TAPI_FEAT_KPI)
                                        "kpi (%4d/%4d), "
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_KPI) */
                                        "evt (%4d/%4d)\n",
                  IFX_TAPI_VoiceBufferPool_ElementAvailCountGet(),
                  IFX_TAPI_VoiceBufferPool_ElementCountGet(),
#if (TAPI_CFG_FEATURES & TAPI_FEAT_KPI)
                  IFX_TAPI_KPI_BufferPool_ElementAvailCountGet(),
                  IFX_TAPI_KPI_VoiceBufferPool_ElementCountGet(),
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_KPI) */
                  IFX_TAPI_EventWrpBufferPool_ElementAvailCountGet(),
                  IFX_TAPI_EventWrpBufferPool_ElementCountGet());

   return len;
}

#ifdef TAPI_PACKET_OWNID
/**
   Print out to the console voice buffer information

   \param  buf          Destination buffer.

   \return
      Always zero.

   \remarks
      That function do not using buffer because output too large for
      **non-circled** buffer with **unknown** size.
*/
static IFX_int32_t proc_read_voice_bufferpool (IFX_char_t *buf)
{
   IFX_TAPI_VoiceBufferPoolStatusShow();

   return 0;
}
#endif /* TAPI_PACKET_OWNID */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS)
/**
   Read the statistic information from the driver.

   \param   buf     Buffer to fill in the data.
   \param   start   Returns the increment to the offset for the next call.
   \param   offset  Offset where to begin output.
   \param   count   Number of characters in buffer.
   \param   eof     Return if eof was encountered.
   \param   data    Unused.
   \return  count   Number of returned characters.
*/
static int proc_read_tapi_stats(char *buf, char **start, off_t offset,
                                int count, int *eof, void *data)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx  = IFX_NULL;
   TAPI_DEV           *pTapiDev = IFX_NULL;
   TAPI_CHANNEL       *pChannel = IFX_NULL;
   IFX_uint16_t       i, j, k;
   int                len=0;

   /* Check all slots for registered drivers */
   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      pDrvCtx = gHLDrvCtx[i].pDrvCtx;

      if (pDrvCtx == IFX_NULL)
      {
         continue;
      }

      if (pDrvCtx->pTapiDev == IFX_NULL)
      {
         /* driver not initialised yet -> skip it */
         break;
      }

      /* Loop over all devices of the registered driver */
      for (j = 0; j < pDrvCtx->maxDevs; j++)
      {
         pTapiDev = &(pDrvCtx->pTapiDev[j]);

         if (!pTapiDev->bInitialized)
         {
            continue;
         }

         if (offset != 0)
         {
            offset--;
         }
         else
         {
            /* Headline */
            len += sprintf(buf+len,
               "-- %s #%d --\n", pDrvCtx->drvName, j);
            len += sprintf(buf+len, "CH STRM "
               "TX:     ok  discarded  congested "
               "RX:     ok  discarded  congested\n");

            /* Loop over all TAPI channels of this device */
            for (k=0; k < pDrvCtx->maxChannels; k++)
            {
               pChannel = &pTapiDev->pTapiChanelArray[k];

               if (k < pChannel->pTapiDevice->nResource.CodCount)
               {
                  len += sprintf(buf+len, "%2d COD  "
                     "%10u %10u %10u "
                     "%10u %10u %10u\n", k,
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_COD,
                                       TAPI_STAT_COUNTER_EGRESS_DELIVERED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_COD,
                                       TAPI_STAT_COUNTER_EGRESS_DISCARDED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_COD,
                                       TAPI_STAT_COUNTER_EGRESS_CONGESTED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_COD,
                                       TAPI_STAT_COUNTER_INGRESS_DELIVERED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_COD,
                                       TAPI_STAT_COUNTER_INGRESS_DISCARDED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_COD,
                                       TAPI_STAT_COUNTER_INGRESS_CONGESTED));
               }

               if (k < pChannel->pTapiDevice->nResource.DectCount)
               {
                  len += sprintf(buf+len, "%2d DECT "
                     "%10u %10u %10u "
                     "%10u %10u %10u\n", k,
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_DECT,
                                       TAPI_STAT_COUNTER_EGRESS_DELIVERED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_DECT,
                                       TAPI_STAT_COUNTER_EGRESS_DISCARDED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_DECT,
                                       TAPI_STAT_COUNTER_EGRESS_CONGESTED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_DECT,
                                       TAPI_STAT_COUNTER_INGRESS_DELIVERED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_DECT,
                                       TAPI_STAT_COUNTER_INGRESS_DISCARDED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_DECT,
                                       TAPI_STAT_COUNTER_INGRESS_CONGESTED));
               }

#if (TAPI_CFG_FEATURES & TAPI_FEAT_HDLC)
               if (k < pChannel->pTapiDevice->nResource.HdlcCount)
               {
                  len += sprintf(buf+len, "%2d HDLC "
                     "%10u %10u %10u "
                     "%10u %10u %10u\n", k,
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_HDLC,
                                       TAPI_STAT_COUNTER_EGRESS_DELIVERED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_HDLC,
                                       TAPI_STAT_COUNTER_EGRESS_DISCARDED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_HDLC,
                                       TAPI_STAT_COUNTER_EGRESS_CONGESTED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_HDLC,
                                       TAPI_STAT_COUNTER_INGRESS_DELIVERED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_HDLC,
                                       TAPI_STAT_COUNTER_INGRESS_DISCARDED),
                     IFX_TAPI_Stat_Get(pChannel, IFX_TAPI_STREAM_HDLC,
                                       TAPI_STAT_COUNTER_INGRESS_CONGESTED));
               }
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_HDLC) */
            }

            *start = (char *)1;
            *eof = 1;
            return len;
         }
      }
   }

   *eof = 0;

   return len;
}


/**
   Reset all the packet path statistic information of the driver.

   \param   file    file structure for proc file
   \param   buffer  buffer holding the data
   \param   count   number of characters in buffer
   \param   data    unused
   \return  count   Number of processed characters
*/
static int proc_write_tapi_stats(struct file *file, const char *buffer,
                                 unsigned long count, void *data)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx  = IFX_NULL;
   TAPI_DEV           *pTapiDev = IFX_NULL;
   TAPI_CHANNEL       *pChannel = IFX_NULL;
   IFX_uint16_t       i, j, k;

   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      pDrvCtx = gHLDrvCtx[i].pDrvCtx;

      if (pDrvCtx == IFX_NULL)
      {
         continue;
      }

      /* Loop over all devices of the registered driver */
      for (j = 0; j < pDrvCtx->maxDevs; j++)
      {
         if (pDrvCtx->pTapiDev == IFX_NULL)
         {
            /* driver not initialised yet -> skip it */
            break;
         }

         pTapiDev = &(pDrvCtx->pTapiDev[j]);

         if (!pTapiDev->bInitialized)
         {
            continue;
         }

         /* Loop over all TAPI channels of this device */
         for (k=0; k < pDrvCtx->maxChannels; k++)
         {
            pChannel = &pTapiDev->pTapiChanelArray[k];
            IFX_TAPI_Stat_Reset(pChannel);
         }
      }
   }

   return count;
}
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION)
/**
   Read the registered low level driver information

   \param  buf          Destination buffer.

   \return
   Number of bytes written into the buffer.
*/
static IFX_int32_t proc_read_tapi_ppd(IFX_char_t *buf)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx = IFX_NULL;
   TAPI_DEV           *pTapiDev = IFX_NULL;
   IFX_int32_t i, j, k, len = 0;

   len += sprintf(buf+len,
      "===========================================================================================================\n");
   len += sprintf(buf+len, "%4s| %3s| %5s| %6s| %18s| %6s| %21s| %6s| %6s| %6s| %6s|\n",
          "",   "", "phone",     "sm",  "",     "active",     "last",   "cap.",      "",       "",       "");
   len += sprintf(buf+len, "%4s| %3s| %5s| %6s| %18s| %6s| %21s| %6s| %6s| %6s| %6s|\n",
       "dev", "ch",  "det.", "status", "state",  "timer", "capacity[nF]", "thresh", "T1[s]", "T2[ms]", "T3[ms]");
   len += sprintf(buf+len,
      "===========================================================================================================\n");

   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      pDrvCtx = gHLDrvCtx[i].pDrvCtx;

      if (pDrvCtx == IFX_NULL)
      {
         continue;
      }

      if (pDrvCtx->pTapiDev == IFX_NULL)
      {
         /* driver not initialised yet -> skip it */
         break;
      }

      /* Loop over all devices of the registered driver */
      for (j = 0; j < pDrvCtx->maxDevs; j++)
      {
         pTapiDev = &(pDrvCtx->pTapiDev[j]);

         if (!pTapiDev->bInitialized)
         {
            continue;
         }
         /* channel status */
         /* for the moment limit printout to two channels until the real
            number of channels is known internally */
         for (k = 0; (k < pTapiDev->nMaxChannel) && (k < 2); k++)
         {
            TAPI_CHANNEL *pChannel = pTapiDev->pTapiChanelArray + k;
            TAPI_PPD_DATA_t    *pTapiPpdData = pChannel->pTapiPpdData;

            if (pTapiPpdData == IFX_NULL)
            {
               continue;
            }
            IFX_TAPI_PPD_proc_read(pChannel, j, buf, &len);
         }
      }
   }

   return len;
}
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION) */


/**
   Read the current fifo status

   \param  buf          Destination buffer.

   \return
   Number of bytes written into the buffer.
*/
static IFX_int32_t proc_read_fifos (IFX_char_t *buf)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx  = IFX_NULL;
   TAPI_DEV           *pTapiDev = IFX_NULL;
   TAPI_CHANNEL       *pChannel = IFX_NULL;
   IFX_int32_t        len = 0;
   IFX_int32_t        i, j, k;
   IFX_TAPI_STREAM_t  nStream;
   IFX_uint32_t       nMaxStream;
   static char        *pStreamName[] = { "COD", "DECT", "HDLC"};
   IFX_TAPI_EVENT_FIFO_COUNTERS_t evtCounters;
   IFX_TAPI_UPSTREAM_FIFO_COUNTERS_t fdCounters;

   /* maximum stream we are going to print */
   nMaxStream = sizeof(pStreamName)/sizeof(*pStreamName);
   if (IFX_TAPI_STREAM_MAX < nMaxStream)
   {
      nMaxStream = IFX_TAPI_STREAM_MAX;
   }

   /* for all LL-drivers */
   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      if (gHLDrvCtx [i].pDrvCtx == IFX_NULL)
      {
         continue;
      }

      pDrvCtx = gHLDrvCtx [i].pDrvCtx;
      len += sprintf(buf+len, "-- %s --\n", pDrvCtx->drvName);

      if (pDrvCtx->pTapiDev == IFX_NULL)
      {
         /* driver not initialised yet -> skip it */
         break;
      }

      /* for all devices */
      for (j = 0; j < pDrvCtx->maxDevs; j++)
      {
         pTapiDev = &(pDrvCtx->pTapiDev[j]);

         /* Headline (once per device) */

         len += sprintf(buf+len, "DEV%2d: evtH evtL ", j);

         len += sprintf(buf+len, " fd:");
         for (nStream = 0; nStream < nMaxStream; nStream++)
         {
            if (pDrvCtx->readChannels[nStream] > 0)
            {
               len += sprintf(buf+len, "%4s ", pStreamName[nStream]);
            }
         }
#if (TAPI_CFG_FEATURES & TAPI_FEAT_KPI)
         len += sprintf(buf+len, " KPI-CH:");
         for (nStream = 0; nStream < nMaxStream; nStream++)
         {
            len += sprintf(buf+len, "%4s ", pStreamName[nStream]);
         }
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_KPI) */

         /* Data line (one per channel) */

         /* Loop over all TAPI channels of this device */
         for (k=0; k < pDrvCtx->maxChannels; k++)
         {
            pChannel = &pTapiDev->pTapiChanelArray[k];

            len += sprintf(buf+len, "\n CH%2d:", k);

            /* Event fifos */
            if (TAPI_statusOk ==
                IFX_TAPI_EventFifoCounters (pChannel, &evtCounters))
            {
               len += sprintf(buf+len, " %4d %4d ",
                              evtCounters.nHighWaiting,
                              evtCounters.nLowWaiting);
            }
            else
            {
               len += sprintf(buf+len, " FAIL FAIL ");
            }

            /* Fifos on file descriptors */
            len += sprintf(buf+len, "    ");
            for (nStream = 0; nStream < nMaxStream; nStream++)
            {
               if (pDrvCtx->readChannels[nStream] == 0)
                  continue;
               if (TAPI_statusOk ==
                   IFX_TAPI_UpStreamFifo_Counters(pChannel, nStream, &fdCounters))
               {
                  len += sprintf(buf+len, "%4d ", fdCounters.nWaiting);
               }
               else
               {
                  len += sprintf(buf+len, "FAIL ");
               }
            }

            /* KPI Channels */
#if (TAPI_CFG_FEATURES & TAPI_FEAT_KPI)
            len += sprintf(buf+len, "        ");
            /* protect channel-data from mutual access */
            TAPI_OS_MutexGet (&pChannel->semTapiChDataLock);
            for (nStream = 0; nStream < nMaxStream; nStream++)
            {
               len += sprintf(buf+len, "%04X ",
                              IFX_TAPI_KPI_ChGet(pChannel, nStream));
            }
            /* release channel-data protection */
            TAPI_OS_MutexRelease (&pChannel->semTapiChDataLock);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_KPI) */
         } /* for all channels */
         len += sprintf(buf+len, "\n");
      } /* for all devices */
   } /* for all LL-drivers */

   return len;
}


/**
   The proc filesystem: function to read an entry.
   This function provides information of proc files to the user space

   \param   page    Buffer to fill in the data.
   \param   start   Returns the increment to the offset for the next call.
   \param   off     Offset where to begin output.
   \param   count   Number of characters in buffer.
   \param   eof     Return if eof was encountered.
   \param   data    Unused.
   \return  Number of returned characters.
*/
static int proc_read_tapi(char *page, char **start,
                          off_t off, int count, int *eof, void *data)
{
    int len;

    IFX_int32_t (*fn)(IFX_char_t *buf);

    /* write data into the page */
    if (data != IFX_NULL)
    {
        fn = data;
        /* FIXME: possible buffer overflow!!!
         * buffer handling have to be circled */
        len = (int)fn((IFX_char_t *)page);
    }
    else
        return 0;

    if (len <= off+count)
        *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count)
        len = count;
    if (len<0)
        len = 0;
    /* return the data length  */
    return len;
}


/**
   Initialize and install the proc entry

   \return
   -1 or 0 on success
*/
/*lint -save -e611
\todo: FIXME
   (ANSI X3.159-1989)
   It is invalid to convert a function pointer to an object pointer
   or a pointer to void, or vice-versa.
*/
static IFX_int32_t proc_EntriesInstall(IFX_void_t)
{
   struct proc_dir_entry *driver_proc_node,
                         *proc_file;

   /* install the proc entry */
   TRACE(TAPI_DRV, DBG_LEVEL_LOW, ("TAPI_DRV: using proc fs\n"));
   driver_proc_node = proc_mkdir( "driver/" DRV_TAPI_NAME, IFX_NULL);
   if (driver_proc_node != IFX_NULL)
   {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      driver_proc_node->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */

      proc_file = create_proc_read_entry( "version" , S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *)proc_get_tapi_version);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */

      proc_file = create_proc_read_entry( "status" , S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *)proc_get_tapi_status);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */

      proc_file = create_proc_read_entry( "registered_drivers" , S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *)proc_get_tapi_registered_drivers);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */

      proc_file = create_proc_read_entry( "bufferpool", S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *) proc_read_bufferpool);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */

#ifdef TAPI_PACKET_OWNID
      proc_file = create_proc_read_entry( "voice_buffers", S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *) proc_read_voice_bufferpool);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */
#endif /* TAPI_PACKET_OWNID */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS)
      {
         struct proc_dir_entry *proc_stat_node;

         if ((proc_stat_node =
                 create_proc_entry ("statistic", 0, driver_proc_node)))
         {
            proc_stat_node->read_proc  = proc_read_tapi_stats;
            proc_stat_node->write_proc = proc_write_tapi_stats;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
            proc_stat_node->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */
         }
      }
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION)
      proc_file = create_proc_read_entry( "phone_detection", S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *) proc_read_tapi_ppd);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION) */

#ifdef HAVE_CONFIG_H
      proc_file = create_proc_read_entry("configure", S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *)proc_ConfigureGet);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */
#endif /* HAVE_CONFIG_H */

      proc_file = create_proc_read_entry( "fifo_status" , S_IFREG|S_IRUGO,
                                         driver_proc_node, proc_read_tapi,
                                         (void *)proc_read_fifos);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
      if (proc_file != NULL)
         proc_file->owner = THIS_MODULE;
#endif /* < Linux 2.6.30 */
   }
   else
   {
      TRACE(TAPI_DRV,DBG_LEVEL_HIGH,("TAPI_DRV: cannot create proc entry\n"));
      return -1;
   }

   return 0;
}
/*lint -restore */


/**
   Remove proc filesystem entries.

   \return
   None.

   \remarks
   Called by the kernel.
*/
static IFX_void_t proc_EntriesRemove(IFX_void_t)
{
   remove_proc_entry("driver/" DRV_TAPI_NAME "/fifo_status",0);
#ifdef HAVE_CONFIG_H
   remove_proc_entry("driver/" DRV_TAPI_NAME "/configure",0);
#endif /* HAVE_CONFIG_H */
#if (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION)
   remove_proc_entry("driver/" DRV_TAPI_NAME "/phone_detection" ,0);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PHONE_DETECTION) */
#if (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS)
   remove_proc_entry("driver/" DRV_TAPI_NAME "/statistic" ,0);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_STATISTICS) */
   remove_proc_entry("driver/" DRV_TAPI_NAME "/version" ,0);
   remove_proc_entry("driver/" DRV_TAPI_NAME "/status",0);
   remove_proc_entry("driver/" DRV_TAPI_NAME "/registered_drivers",0);
   remove_proc_entry("driver/" DRV_TAPI_NAME "/bufferpool",0);
#ifdef TAPI_PACKET_OWNID
   remove_proc_entry("driver/" DRV_TAPI_NAME "/voice_buffers",0);
#endif /* TAPI_PACKET_OWNID */
   remove_proc_entry("driver/" DRV_TAPI_NAME ,0);

   return;
}
#endif /* CONFIG_PROC_FS */


#ifdef LINUX_2_6
/**
   Set the scheduling policy for the TAPIevent workqueue to RT scheduling

   \param  foo          unused.
*/
static IFX_void_t tapi_wq_setscheduler (IFX_int32_t foo)
{
   struct sched_param sched_params;

   sched_params.sched_priority = TAPI_OS_THREAD_PRIO_HIGH;
   sched_setscheduler(current, SCHED_FIFO, &sched_params);
}
#endif /* LINUX_2_6 */


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static IFX_int32_t tapiClassCreate(IFX_void_t)
{
   pTAPI_Class = class_create(THIS_MODULE, "tapi");
   if (IS_ERR(pTAPI_Class))
   {
      TRACE( TAPI_DRV, DBG_LEVEL_HIGH,
         ("tapiClassCreate: unable to create tapi class\n"));
      return TAPI_statusErr;
   }
   return 0;
}

static IFX_void_t tapiClassRemove(IFX_void_t)
{
   class_destroy(pTAPI_Class);
}
#endif /* >= Linux 2.6.0 */

/**
   Initialize the module.

   \return
   Error code or 0 on success

   \remarks
   Called by the kernel.
*/
static int __init ifx_tapi_module_init(void)
{
   /* copyrigth trace shall not be prefixed with KERN_INFO */
   printk("%s, (c) 2001-2010 Lantiq Deutschland GmbH\n", &TAPI_WHATVERSION[4]);

   if (IFX_TAPI_Driver_Start() != IFX_SUCCESS)
   {
      printk (KERN_ERR "Driver start failed\n");
      return -1;
   }

   memset (&tapi_fops, 0, sizeof (tapi_fops));

#ifdef CONFIG_PROC_FS
   proc_EntriesInstall();
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
   if (tapiClassCreate() != IFX_SUCCESS)
      return -1;
#endif /* >= Linux 2.6.0 */


#ifdef LINUX_2_6
   pTapiWq = create_workqueue("TAPIevents");
   TAPI_DeferWork((IFX_void_t *) tapi_wq_setscheduler,
                  (IFX_void_t *)&tapi_wq_setscheduler_param);
#endif /* LINUX_2_6 */

   return 0;
}

/**
   Clean up the module.

   \remarks
   Called by the kernel.
*/
static void __exit ifx_tapi_module_exit(void)
{
   int i;

   printk (KERN_INFO "Removing Highlevel TAPI module\n");

   /* Actually all LL drivers should have unregistered here. Being careful
      we force unregister of any drivers which may still be registered. */
   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      if (gHLDrvCtx [i].pDrvCtx != IFX_NULL)
      {
         IFX_TAPI_Unregister_LL_Drv (gHLDrvCtx [i].pDrvCtx->majorNumber);
      }
   }

#ifdef CONFIG_PROC_FS
   proc_EntriesRemove();
#endif /* CONFIG_PROC_FS */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
   tapiClassRemove();
#endif /* >= Linux 2.6.0 */

#ifdef LINUX_2_6
   /* as we are using work queues to schedule events from the interrupt
      context to the process context, we use work queues in case of
      Linux 2.6. they must be flushed on driver unload... */
   flush_workqueue(pTapiWq);
   destroy_workqueue(pTapiWq);
#endif /* LINUX_2_6 */

   IFX_TAPI_Driver_Stop();

   TRACE(TAPI_DRV, DBG_LEVEL_NORMAL, ("TAPI_DRV: cleanup successful\n"));
}
#endif /* IFXOS_USE_DEV_IO */

#ifdef TAPI_HAVE_TIMERS

#ifndef __KERNEL__
/**
   Timer thread function for processing user space timers

   \param  pThread  thread handler

   \return

   \remarks
   Thread started for each timer, and can be stopped using
   timer stop event linked to this thread.
*/
static IFX_int32_t TAPI_Timer_thread (IFXOS_ThreadParams_t *pThread)
{
   Timer_ID Timer = (Timer_ID)pThread->nArg1;
   IFX_uint32_t ret;
   if (Timer == IFX_NULL)
      return IFX_FALSE;

   if (Timer->Periodical_Time == 0)
   {
      Timer->pTimerEntry (Timer, Timer->nArgument);
      return IFX_SUCCESS;
   }
   while (IFX_TRUE)
   {
      TAPI_OS_EventWait (&Timer->TapiTimerStopEvent,
                         Timer->Periodical_Time, &ret);
      if (ret == 0)
      {
         /* stop timer event received, exit thread */
         break;
      }
      /* Enter call back function */
      Timer->pTimerEntry (Timer, Timer->nArgument);
      if (!Timer->bPeriodical)
      {
         /* Oneshot timer expired, exit thread */
         break;
      }
   }
   return IFX_SUCCESS;
}
#endif /* __KERNEL__ */

/**
   Function create a timer.

   \param  pTimerEntry  Functionpointer to the call back function.
   \param  nArgument    Pointer to TAPI channel structure.

   \return
   Timer_ID    - pointer to internal timer structure

   \remarks
   Initialize a task queue which will be scheduled once a timer interrupt occurs
   to execute the appropriate operation in a process context, process in which
   semaphores ... are allowed.
   Please notice that this task has to run under the keventd process, in which it
   can be executed thousands of times within a single timer tick.
*/
Timer_ID TAPI_Create_Timer(TIMER_ENTRY pTimerEntry, IFX_ulong_t nArgument)
{
#ifdef __KERNEL__
   Timer_ID pTimerData;

   /* allocate memory for the channel */
   pTimerData = kmalloc(sizeof (*pTimerData), GFP_KERNEL);
   if (pTimerData == IFX_NULL)
      return IFX_NULL;

   /* set function to be called after timer expires */
   pTimerData->pTimerEntry = pTimerEntry;
   pTimerData->nArgument = nArgument;
   pTimerData->bStopped = IFX_FALSE;

   init_timer(&(pTimerData->Timer_List));

   /* set timer call back function */
   pTimerData->Timer_List.function = TAPI_timer_call_back;
   pTimerData->Timer_List.data = (IFX_ulong_t) pTimerData;

   /* Initialize Timer Task */
#ifdef LINUX_2_6
   #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
   INIT_WORK(&(pTimerData->timerTask), TAPI_tqueue, (IFX_void_t *) pTimerData);
   #else
   INIT_WORK(&(pTimerData->timerTask), TAPI_tqueue);
   #endif
#else
   /* initialize tq_struct member of Timer_ID structure */
   memset(&(pTimerData->timerTask), 0, sizeof (struct tq_struct));
   pTimerData->timerTask.routine = TAPI_tqueue;
   pTimerData->timerTask.data = (IFX_void_t *) pTimerData;
#endif /* LINUX_2_6 */
   return pTimerData;
#else /* ! __KERNEL__ */
   Timer_ID pTimerData;
   pTimerData = (Timer_ID)TAPI_OS_Malloc(sizeof (*pTimerData));
   if (pTimerData == IFX_NULL)
      return IFX_NULL;
   TAPI_OS_LockGet (&pTimerData->LockTimerEvent);
   TAPI_OS_EventInit (&pTimerData->TapiTimerStopEvent);
   TAPI_OS_LockRelease (&pTimerData->LockTimerEvent);
   /* set function to be called after timer expires */
   pTimerData->pTimerEntry = pTimerEntry;
   pTimerData->nArgument = nArgument;
   return pTimerData;
#endif /* __KERNEL__ */
}

/**
   Function set and starts a timer with a specific time. It can be choose if the
   timer starts periodically.

   \param  Timer_ID     Pointer to internal timer structure.
   \param  nTime        Time in ms.
   \param  bPeriodically Starts the timer periodically or not.
   \param  bRestart     Restart the timer or normal start.

   \return
   Returns an error code: IFX_TRUE / IFX_FALSE
*/
IFX_boolean_t TAPI_SetTime_Timer(Timer_ID Timer, IFX_uint32_t nTime,
                                 IFX_boolean_t bPeriodically,
                                 IFX_boolean_t bRestart)
{
   if (Timer == 0)
   {
     TRACE(TAPI_DRV,DBG_LEVEL_HIGH,
          ("\n\rDRV_ERROR: TAPI_SetTime_Timer (Timer == 0)\n\r"));
     return IFX_FALSE;
   }
#ifdef __KERNEL__
   Timer->Periodical_Time = HZ*nTime/1000;
   Timer->bPeriodical = bPeriodically;
   Timer->bStopped = IFX_FALSE;

   if (timer_pending(&(Timer->Timer_List)))
   {
      if (IFX_FALSE == bRestart)
         return IFX_TRUE;

      if (mod_timer(&(Timer->Timer_List), jiffies + Timer->Periodical_Time))
         return IFX_TRUE;
   }

   Timer->Timer_List.expires = jiffies + Timer->Periodical_Time;
   add_timer(&(Timer->Timer_List));
#else /* !__KERNEL__ */
   Timer->TimerThread.bValid = IFX_FALSE;
   Timer->bPeriodical = bPeriodically;
   Timer->Periodical_Time = nTime;
   if (bRestart == IFX_TRUE)
   {
      TAPI_Stop_Timer (Timer);
   }

   TAPI_OS_LockGet (&Timer->LockTimerEvent);
   TAPI_OS_EventDelete (&Timer->TapiTimerStopEvent);
   TAPI_OS_EventInit (&Timer->TapiTimerStopEvent);
   TAPI_OS_LockRelease (&Timer->LockTimerEvent);

   TAPI_OS_ThreadInit (&Timer->TimerThread, "TAPItimer",
                       TAPI_Timer_thread,
                       0, TAPI_OS_THREAD_PRIO_HIGHEST, (IFX_ulong_t)Timer, 0);

#endif /* __KERNEL__ */
   return IFX_TRUE;
}

/**
   Function stop a timer.

   \param  Timer_ID     Pointer to internal timer structure.

   \return
   Returns an error code: IFX_TRUE / IFX_FALSE
*/
IFX_boolean_t TAPI_Stop_Timer(Timer_ID Timer)
{
#ifdef __KERNEL__
   /* stop timer */
   Timer->bStopped = IFX_TRUE;
   /* prevent restart of driver */
   Timer->bPeriodical = IFX_FALSE;

   if (timer_pending(&(Timer->Timer_List)))
   {
      /* remove driver from list */
      del_timer_sync(&(Timer->Timer_List));
   }
   return IFX_TRUE;
#else /* ! __KERNEL__ */
   if ((Timer != IFX_NULL) )
   {
      /* Just send stop event, and exit timer thread */
      TAPI_OS_LockGet (&Timer->LockTimerEvent);
      TAPI_OS_EventWakeUp (&Timer->TapiTimerStopEvent);
      TAPI_OS_LockRelease (&Timer->LockTimerEvent);
   }
   return IFX_TRUE;
#endif /* __KERNEL__ */
}

/**
   Function delete a timer.

   \param  Timer_ID     Pointer to internal timer structure.

   \return
   Returns an error code: IFX_TRUE / IFX_FALSE
*/
IFX_boolean_t TAPI_Delete_Timer(Timer_ID Timer)
{
#ifdef __KERNEL__
   if (Timer)
   {
      TAPI_Stop_Timer(Timer);
      /* free memory */
      kfree(Timer);
      return IFX_TRUE;
   }
#else
   if (Timer)
   {
      TAPI_Stop_Timer(Timer);
      TAPI_OS_LockGet (&Timer->LockTimerEvent);
      TAPI_OS_EventDelete (&Timer->TapiTimerStopEvent);
      TAPI_OS_LockRelease (&Timer->LockTimerEvent);

      /* free memory */
      TAPI_OS_Free (Timer);
      return IFX_TRUE;
   }
#endif /* __KERNEL__ */
   return IFX_FALSE;
}
#endif /* TAPI_HAVE_TIMERS */

#ifdef __KERNEL__
/**
   Helper function to get a periodical timer.

   \param  pWork        Pointer to corresponding timer ID.

   \remarks
   This function will be executed in  the process context, so to avoid
   scheduling in Interrupt Mode while working with semaphores etc...
   The task is always running under the keventd process and is also running
   very quickly. Even on a very heavily loaded system, the latency in the
   scheduler queue is quite small
*/
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static IFX_void_t TAPI_tqueue (IFX_void_t *pWork)
#else /* for Kernel newer or equal 2.6.20 */
static IFX_void_t TAPI_tqueue (struct work_struct *pWork)
#endif /* (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)) */
{
   Timer_ID Timer = (Timer_ID) pWork;

   if (Timer->bStopped)
      return;

   /* Call TAPI Timer function */
   Timer->pTimerEntry(Timer, Timer->nArgument);
   if (Timer->bPeriodical)
   {
      if (timer_pending(&(Timer->Timer_List)))
      {
         /* update existed timer */
         if (mod_timer(&(Timer->Timer_List), jiffies + Timer->Periodical_Time))
            return;
      }

      /* cleate new timer */
      Timer->Timer_List.expires = jiffies + Timer->Periodical_Time;
      add_timer(&(Timer->Timer_List));
   }
}

#ifdef __KERNEL__
/**
   Helper function to get a periodical timer.

   \param  arg          Pointer to corresponding timer ID.
*/
static IFX_void_t TAPI_timer_call_back (IFX_ulong_t arg)
{
   Timer_ID Timer = (Timer_ID) arg;
   /* do the operation in process context,
      not in interrupt context */
#ifdef LINUX_2_6
   queue_work (pTapiWq, &(Timer->timerTask));
#else
   schedule_task (&(Timer->timerTask));
#endif /* LINUX_2_6 */
}
#endif /* __KERNEL__ */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
/* High Level Event Dispatcher function. */
static IFX_void_t Deferred_Worker (IFX_void_t *pWork)
{
   IFX_TAPI_EXT_EVENT_PARAM_t *pEvParam = (IFX_TAPI_EXT_EVENT_PARAM_t *) pWork;
   pEvParam->pFunc(pEvParam);
}
#else /* for Kernel newer or equal 2.6.20 */
static IFX_void_t Deferred_Worker (struct work_struct *pWork)
{
   IFX_TAPI_EXT_EVENT_PARAM_t *pEvParam = (IFX_TAPI_EXT_EVENT_PARAM_t *) pWork;
   pEvParam->pFunc(pEvParam);
}
#endif

/**
   Defer work to process context

   \param  pFunc        Pointer to function to be called.
   \param  pParam       Parameter passed to the function.
                        Attention: this must be a valid structure of
                        type IFX_TAPI_EXT_EVENT_PARAM_t as we need to
                        do some casting for Linux.

   \return TAPI_statusOk or TAPI_statusErr in case of an error.
*/
IFX_int32_t TAPI_DeferWork (IFX_void_t *pFunc, IFX_void_t *pParam)
{
   IFX_int32_t ret = TAPI_statusOk;
   IFX_TAPI_EXT_EVENT_PARAM_t *pEvParam = (IFX_TAPI_EXT_EVENT_PARAM_t *) pParam;
#ifdef LINUX_2_6
   struct work_struct         *pTapiWs;

   pTapiWs = (struct work_struct *) &pEvParam->tapiWs;
   pEvParam->pFunc   = (IFX_void_t *)pFunc;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
   INIT_WORK(pTapiWs, Deferred_Worker, (IFX_void_t *) pEvParam);
#else
   INIT_WORK(pTapiWs, Deferred_Worker);
#endif

   if (queue_work (pTapiWq, pTapiWs) == 0)
   {
      ret = TAPI_statusWorkFail;
   }
#else
   struct tq_struct           *pTapiTq;

   pTapiTq           = (struct tq_struct *)&pEvParam->tapiTq;
   INIT_TQUEUE(pTapiTq, Deferred_Worker, (IFX_void_t *)pEvParam);
   pEvParam->pFunc   = (IFX_void_t *)pFunc;
   if (schedule_task (pTapiTq) == 0)
   {
      ret = TAPI_statusWorkFail;
   }
#endif /* LINUX_2_6 */
   return ret;
}

/**
   Executes the select for the channel fd

   \param  pTapiPriv    Pointer fd private data carrying the context.
   \param  pNode        node list.
   \param  pOpt         Optional argument, which contains needed information for
                        TAPI_OS_DrvSelectQueueAddTask.

   \return
   System event qualifier. Either 0 or TAPI_OS_SYSREAD

   \remarks
   This function needs operating system services, that are hidden by
   IFXOS macros.
*/
static IFX_int32_t TAPI_SelectCh (TAPI_FD_PRIV_DATA_t *pTapiPriv,
                                  TAPI_OS_drvSelectTable_t *pNode,
                                  TAPI_OS_drvSelectOSArg_t *pOpt)
{
   TAPI_CHANNEL *pTapiCh = (TAPI_CHANNEL* )pTapiPriv->pTapiCtx;
   IFX_TAPI_DRV_CTX_t *pDrvCtx = (IFX_TAPI_DRV_CTX_t*)
                                 pTapiCh->pTapiDevice->pDevDrvCtx;
   IFX_TAPI_LL_DEV_t  *pLLDev = pTapiCh->pTapiDevice->pLLDev;
   IFX_uint32_t  flags = 0;
   IFX_int32_t   ret = 0;
#if (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38)
   IFX_TAPI_T38_STATUS_t TapiFaxStatus;
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38)
  memset (&TapiFaxStatus, 0, sizeof (IFX_TAPI_T38_STATUS_t));

   /* Get the Status from the low level driver */
   if (IFX_TAPI_PtrChk (pDrvCtx->COD.T38_Status_Get))
      (IFX_void_t)pDrvCtx->COD.T38_Status_Get (pTapiCh->pLLChannel, &TapiFaxStatus);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38) */

   /* Register the voice channel waitqueues as wakeup source. */
   TAPI_OS_DrvSelectQueueAddTask (pOpt, &pTapiCh->semReadBlock.object, pNode);
   TAPI_OS_DrvSelectQueueAddTask (pOpt, &pTapiCh->wqRead, pNode);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38)
   TAPI_OS_DrvSelectQueueAddTask (pOpt, &pTapiCh->wqWrite, pNode);

   if ((TapiFaxStatus.nStatus & IFX_TAPI_FAX_T38_TX_ON) &&
       (pTapiCh->bFaxDataRequest == IFX_TRUE))
   {
      /* task should write a new packet now */
      ret |= TAPI_OS_SYSWRITE;
   }
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_FAX_T38) */

   /* select on a voice channel -- only implemented for TAPI */
   flags |= CF_NEED_WAKEUP;
   /* clear flags first, then apply new flags */

   if (pDrvCtx->IRQ.LockDevice != IFX_NULL)
   {
      pDrvCtx->IRQ.LockDevice (pLLDev);
   }

#if (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET)
   if ( (pTapiCh->pUpStreamFifo[pTapiPriv->fifo_idx] != IFX_NULL) &&
        !fifoEmpty(pTapiCh->pUpStreamFifo[pTapiPriv->fifo_idx]) )
   {
      flags |= CF_WAKEUPSRC_STREAM;
      flags &= ~CF_NEED_WAKEUP;
      ret |= TAPI_OS_SYSREAD;
   }
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET) */
   pTapiCh->nFlags &= ~(CF_WAKEUPSRC_GR909 | CF_WAKEUPSRC_STREAM |
                        CF_WAKEUPSRC_TAPI | CF_NEED_WAKEUP);
   pTapiCh->nFlags |= flags;

   if (pDrvCtx->IRQ.UnlockDevice != IFX_NULL)
   {
      pDrvCtx->IRQ.UnlockDevice (pLLDev);
   }

   return ret;
}


/**
   Modify own thread priority.

   \param  newPriority  New thread priority.

   \return
   - IFX_SUCCESS priority changed.
   - IFX_ERROR priority not changed.
*/
IFX_int32_t TAPI_OS_ThreadPriorityModify(IFX_uint32_t newPriority)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
#if 0
   ret = setscheduler(current->pid, SCHED_FIFO, &sched_params);
#else /* 0 */
   current->policy = SCHED_FIFO;
   current->rt_priority = newPriority;
#endif /* 0 */
#else
   struct sched_param sched_params;
   IFX_int32_t ret = 0;

   sched_params.sched_priority = newPriority;

   ret = sched_setscheduler(current, SCHED_FIFO, &sched_params);

   if (ret < 0)
   {
      printk(KERN_ERR "Failed to set the thread priority to %d"
             " ret=%d\n", newPriority, ret);

      return IFX_ERROR;
   }
#endif

   return IFX_SUCCESS;
}


/**
   stop a kernel thread. Called by the removing instance

   \param  pThrCntrl    Pointer to a thread control struct.
*/
IFX_void_t TAPI_OS_ThreadKill(TAPI_OS_ThreadCtrl_t *pThrCntrl)
{
   if ((pThrCntrl) &&
       (TAPI_OS_THREAD_INIT_VALID(pThrCntrl) == IFX_TRUE) &&
       (pThrCntrl->thrParams.bRunning == 1))
   {
         /* trigger user thread routine to shutdown */ \
         pThrCntrl->thrParams.bShutDown = IFX_TRUE;

         /* this function needs to be protected with the big
            kernel lock (lock_kernel()). The lock must be
            grabbed before changing the terminate
            flag and released after the down() call. */
         lock_kernel();
         mb();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
         kill_proc(pThrCntrl->tid, SIGKILL, 1);
#else
         kill_pid(find_vpid(pThrCntrl->tid), SIGKILL, 1);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28) */
         /* release the big kernel lock */
         unlock_kernel();
         wait_for_completion (&pThrCntrl->thrCompletion);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
         /* Now we are sure the thread is in zombie state.
            We notify keventd to clean the process up. */
         kill_proc(2, SIGCHLD, 1);
#endif /* kernel 2.6.23 */
   }
}

/**
   Open API to be called from kernel space.

   \param  name      Pointer to a dev node name (character string)

   \return
      IFX_TAPI_KIO_HANDLE  handler
*/
IFX_TAPI_KIO_HANDLE ifx_tapi_kopen(char *name)
{
   IFX_TAPI_Ch_KernRef_t *pRef;
   IFX_char_t *p, *p1, dev_name[8];
   IFX_int_t i;
   IFX_uint8_t ucMajor = 0;
   IFX_uint8_t ucMinor = 0;
   IFX_char_t *expected_dev_names[MAX_EXPECTED_LL_DRV] =
                                 {"vmmc", "vin", "dxt", "svip", "vxt"};

   /* sanity check */
   for (i=0; i<MAX_EXPECTED_LL_DRV; i++)
   {
      if ((p = strstr (name, expected_dev_names[i])) != IFX_NULL)
         break;
   }

   if (i == MAX_EXPECTED_LL_DRV)
   {
      /* device not found among expected names */
      return IFX_TAPI_KIO_FD_INVALID;
   }

   if (!strncmp (p, "dxt", 3))
   {
      /* patch for DuSLIC-xT dev name */
      strcpy (dev_name, "duslic");
   }
   else
   {
      strcpy (dev_name, expected_dev_names[i]);
   }

   /* extract minor number from name */
   p += strlen(expected_dev_names[i]);
   if (*p)
   {
      ucMinor = simple_strtol(p, &p1, 10);
   }

   /* find a proper major number */
   for (i = 0; i < TAPI_MAX_LL_DRIVERS; i++)
   {
      if (gHLDrvCtx[i].pDrvCtx != IFX_NULL &&
          IFX_NULL != strstr ((gHLDrvCtx[i].pDrvCtx)->drvName, dev_name))
      {
         ucMajor = (gHLDrvCtx[i].pDrvCtx)->majorNumber;
         break;
      }
   }

   if (i == TAPI_MAX_LL_DRIVERS && !ucMajor)
   {
      /* major number not found */
      return IFX_TAPI_KIO_FD_INVALID;
   }

   pRef = TAPI_OS_Malloc (sizeof(*pRef));
   if (IFX_NULL == pRef)
   {
      /* malloc error */
      return IFX_TAPI_KIO_FD_INVALID;
   }

   memset (pRef, 0, sizeof(*pRef));

   pRef->inode.i_rdev = MKDEV (ucMajor, ucMinor);

   if (0 != ifx_tapi_open (&pRef->inode, &pRef->file))
   {
      TAPI_OS_Free (pRef);
      return IFX_TAPI_KIO_FD_INVALID;
   }

   return (IFX_TAPI_KIO_HANDLE)pRef;
}

/**
   Close API to be called from kernel space.

   \param  handle      IFX_TAPI_KIO_HANDLE type

   \return
   0 - if no error,
   otherwise error code
*/
int ifx_tapi_kclose(IFX_TAPI_KIO_HANDLE handle)
{
   int ret = 0;
   IFX_TAPI_Ch_KernRef_t *pRef = (IFX_TAPI_Ch_KernRef_t *)handle;

   if (pRef == IFX_NULL)
   {
      return -ENODEV;
   }

   ret = ifx_tapi_release (&pRef->inode, &pRef->file);
   if (0 != ret)
      return ret;

   TAPI_OS_Free (pRef);

   return 0;
}

/**
   Ioctl API to be called from kernel space.

   \param  handle      IFX_TAPI_KIO_HANDLE type
   \param  command     Ioctl command
   \param  arg         Ioctl argument

   \return
      IFX_SUCCESS or IFX_ERROR
*/
int ifx_tapi_kioctl(IFX_TAPI_KIO_HANDLE handle, unsigned int command, void *arg)
{
   IFX_TAPI_DRV_CTX_t *pDrvCtx = IFX_NULL;
   IFX_TAPI_ioctlCtx_t ctx;
   IFX_int32_t ret;
   struct inode *inode = &((IFX_TAPI_Ch_KernRef_t *)handle)->inode;

   switch (command) {
      case IFX_TAPI_CID_TX_SEQ_START:
         /*lint -fallthrough*/
      case IFX_TAPI_CID_TX_INFO_START:
         /*lint -fallthrough*/
      case IFX_TAPI_CID_CFG_SET:
         /*lint -fallthrough*/
      case IFX_TAPI_EVENT_MULTI_ENABLE:
         /*lint -fallthrough*/
      case IFX_TAPI_EVENT_MULTI_DISABLE:
         /*lint -fallthrough*/
      case IFX_TAPI_CAP_LIST:
         /* that services require nested copy from/to user-space,
            which are not supported for KIO interface */
         return IFX_ERROR;
      default:
         break;
   }

   /* get the device driver context */
   pDrvCtx = IFX_TAPI_DeviceDriverContextGet (MAJOR(inode->i_rdev));
   if (pDrvCtx == IFX_NULL)
   {
      return IFX_ERROR;
   }

   /* get the ioctl context: channel, device etc. */
   ret = TAPI_ioctlContextGet (pDrvCtx, MINOR(inode->i_rdev),
      command, (IFX_ulong_t)arg, IFX_FALSE, &ctx);
   if (IFX_SUCCESS == ret)
   {
      ret = TAPI_Ioctl (&ctx);
   }

   TAPI_ioctlContextPut ((IFX_ulong_t)arg, &ctx);

   return ret;
}


/*lint -save
-e611 -e546 -e19
-esym(752, init_module -e528
*/
module_init (ifx_tapi_module_init);
module_exit (ifx_tapi_module_exit);

/****************************************************************************/

MODULE_AUTHOR           ("Lantiq Deutschland GmbH");
MODULE_DESCRIPTION      ("TAPI Driver - www.lantiq.com");
MODULE_SUPPORTED_DEVICE ("TAPI DEVICE");
MODULE_LICENSE          ("Dual BSD/GPL");

EXPORT_SYMBOL (IFX_TAPI_Register_LL_Drv);
EXPORT_SYMBOL (IFX_TAPI_Unregister_LL_Drv);
EXPORT_SYMBOL (IFX_TAPI_DeviceReset);
EXPORT_SYMBOL (IFX_TAPI_ReportResources);

EXPORT_SYMBOL (IFX_TAPI_Event_DispatchExt);
EXPORT_SYMBOL (IFX_TAPI_Event_Dispatch);
/*
available as macros
EXPORT_SYMBOL (IFX_TAPI_Event_DeferredDispatch);
EXPORT_SYMBOL (IFX_TAPI_Event_ImmediateDispatch);
*/

#if (TAPI_CFG_FEATURES & TAPI_FEAT_CID)
EXPORT_SYMBOL (TAPI_Phone_GetCidRxBuf);
EXPORT_SYMBOL (TAPI_Cid_Abort);
EXPORT_SYMBOL (TAPI_Cid_IsActive);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_CID) */

EXPORT_SYMBOL (TAPI_Create_Timer);
EXPORT_SYMBOL (TAPI_SetTime_Timer);
EXPORT_SYMBOL (TAPI_Stop_Timer);
EXPORT_SYMBOL (TAPI_Delete_Timer);

#ifdef TAPI_VERSION3
EXPORT_SYMBOL (TAPI_Tone_Set_Source);
#endif /* TAPI_VERSION3 */
EXPORT_SYMBOL (TAPI_ToneState);

EXPORT_SYMBOL(fifoInit);
EXPORT_SYMBOL(fifoReset);
EXPORT_SYMBOL(fifoPut);
EXPORT_SYMBOL(fifoGet);
EXPORT_SYMBOL(fifoPeek);
EXPORT_SYMBOL(fifoEmpty);
EXPORT_SYMBOL(fifoFree);
EXPORT_SYMBOL(fifoElements);
EXPORT_SYMBOL(fifoSize);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_KPI)
EXPORT_SYMBOL(IFX_TAPI_KPI_WaitForData);
EXPORT_SYMBOL(IFX_TAPI_KPI_ReadData);
EXPORT_SYMBOL(IFX_TAPI_KPI_WriteData);
EXPORT_SYMBOL(IFX_TAPI_KPI_ChGet);
EXPORT_SYMBOL(IFX_TAPI_KPI_EgressTaskletRegister);
EXPORT_SYMBOL(irq_IFX_TAPI_KPI_PutToEgress);
EXPORT_SYMBOL(IFX_TAPI_KPI_ScheduleIngressHandling);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_KPI) */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET)
EXPORT_SYMBOL(IFX_TAPI_VoiceBufferGet);
EXPORT_SYMBOL(IFX_TAPI_VoiceBufferPut);
EXPORT_SYMBOL(IFX_TAPI_UpStreamFifo_Put);
EXPORT_SYMBOL(IFX_TAPI_UpStreamFifo_Reset);
EXPORT_SYMBOL(IFX_TAPI_VoiceBufferGetWithOwnerId);
EXPORT_SYMBOL(IFX_TAPI_VoiceBufferFreeAllOwnerId);
#ifdef TAPI_PACKET_OWNID
EXPORT_SYMBOL(IFX_TAPI_VoiceBufferChOwn);
#endif /* TAPI_PACKET_OWNID */
#if (TAPI_CFG_FEATURES & TAPI_FEAT_POLL)
EXPORT_SYMBOL(IFX_TAPI_DownStreamFifo_Handle_Get);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_POLL) */
EXPORT_SYMBOL(bufferPoolInit);
EXPORT_SYMBOL(bufferPoolGet);
EXPORT_SYMBOL(bufferPoolPut);
EXPORT_SYMBOL(bufferPoolGrowthLimitSet);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_PACKET) */

/* packet statistic related exports */
EXPORT_SYMBOL(IFX_TAPI_Stat_Add);

/* FXO related exports */
EXPORT_SYMBOL(IFX_TAPI_Register_DAA_Drv);
EXPORT_SYMBOL(IFX_TAPI_FXO_Event_Dispatch);

#if (TAPI_CFG_FEATURES & TAPI_FEAT_QOS)
EXPORT_SYMBOL(IFX_TAPI_QOS_DrvRegister);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_QOS) */
#endif /* __KERNEL__ */

#if (TAPI_CFG_FEATURES & TAPI_FEAT_FXO)
EXPORT_SYMBOL(IFX_TAPI_Update_SlicFxo);
#endif /* (TAPI_CFG_FEATURES & TAPI_FEAT_FXO) */

EXPORT_SYMBOL (ifx_tapi_kopen);
EXPORT_SYMBOL (ifx_tapi_kclose);
EXPORT_SYMBOL (ifx_tapi_kioctl);
/*lint -restore*/
